"""Set of functions to read .nid AFM data.
This is a first prototype, format is not standard."""
## read_nid is OBSOLETE -> use format_reader.read_nid

import numpy as npoints
from configparser import ConfigParser
import io
import struct
import numpy as np
import logging
from configparser import NoOptionError
from dataIO.config.make_config import string_to_config
import os 


""" dataset info     
Version=2
Points=256
Lines=256
Frame=Scan forward
CurLine=255
Dim0Name=X*
Dim0Unit=m
Dim0Range=1e-05
Dim0Min=0
Dim1Name=Y*
Dim1Unit=m
Dim1Range=1e-05
Dim1Min=0
Dim2Name=Z-Axis
Dim2Unit=m
Dim2Range=1.04e-05
Dim2Min=-5.2e-06
SaveMode=Binary
SaveBits=32
SaveSign=Signed
SaveOrder=Intel
"""

def read_raw_nid (file_name):
    """return header and raw data (all data blocks together) from file_name"""
   
    with open(file_name, 'rb') as binfile:
        lines = binfile.readlines()

    header=[]
    for i,l in enumerate(lines):
        text = l.decode('ansi')
        if '#!' in text:
            break
        header.append(text.strip())

    with open(file_name, 'rb') as binfile:
        a=binfile.read()
    i = a.find(b'#!')

    data = a[i+2:]
    
    return header,data

def read_datablock(data, npoints, nlines, nbits, nim=0, fmt = '<l'):
    """read image of index `nim` from binary data.
    
    `nbits` is redundant, can be obtained from fmt.
    fmt is c-style format string,
    see https://docs.python.org/3.5/library/struct.html#struct-format-strings for other formats
    """
    if nbits%8 != 0: raise ValueError
    imsize=npoints*nlines*nbits//8
    buf2 = io.BytesIO(data[nim*imsize:(nim+1)*imsize])
    idata = struct.iter_unpack(fmt,buf2.read())
    a=list(idata)
    img = np.array(a).reshape((nlines,npoints))
    return img

def make_channel_tags(meta):
    """ builds list of fixed format string with channel identifiers.
    results can be used to extract single channel data and metadata, e.g. with:
       
    """
    config = string_to_config(meta)
    # create itag, a (ordered) list of frame keys
    itag=[]
    ngroups = config.get('DataSet','GroupCount')
    for g in range(int(ngroups)):    
        grcount = config.get('DataSet','Gr%i-Count'%g )   
        for c in range(1,int(grcount)+1):     
            cgtag = 'Gr%i-Ch%i'%(g,c)    
            itag.append(cgtag)
    return itag
        

def read_nid(file_name):
    """ 2021/07/02 OBSOLETE: This is function used in ICSO2020.
    It is replaced with standardize version in `format_reader`.
    To convert the old calls to the new:
     
    
    read a file nid. Return a dictionary with keys `'Gr%i-Ch%i'%(g,c)`
    and `data,x,y` as values, one for each image included in nid file.
    Info from header are used to adjust scales of data.
    
    `read_raw_nid` reads `header` as list of strings and `data` as
    single binary block. `read_datablock` is used to extract an image
    from the datablock. `header` can be used by converting it to `config`
    object and reading fields."""
    
    print ("Obsolete, replace read_nid with version in format_reader")
    header, data = read_raw_nid(file_name) 
    
    # build a config object `config` by merging the string.
    from configparser import NoOptionError
    from dataIO.config.make_config import string_to_config
    config = string_to_config(header)
    
    # all columns of the matrix 
    ngroups = config.get('DataSet','GroupCount') #number of groups 
    imgdic={}
    i=0
    logging.info('reading '+file_name)
    #print('reading '+file_name)
    for g in range(int(ngroups)):    
        grcount = config.get('DataSet','Gr%i-Count'%g )
        # all raws of the actual column 
        for c in range(int(grcount)):     
            cgtag = 'Gr%i-Ch%i'%(g,c)
            try:
                datatag = config.get('DataSet',cgtag)
                # then read the header, specially �Points� and �Lines� 
                #ReadDataSetInfo(datatag)    
                npoints = int(config.get(datatag,'Points'))
                nlines = int(config.get(datatag,'Lines' ))
                nbits = int(config.get(datatag,'SaveBits'))
                sign = config.get(datatag,'SaveSign')
                if (int(nbits) != 32) or sign !='Signed':
                    raise ValueError
                else:
                    fmt = '<l'    
                img = read_datablock(data, npoints, nlines, nbits)
                
                #npoints nlines might be inverted
                Dim0Range = float(config.get(datatag,'Dim0Range'))
                Dim0Min = float(config.get(datatag,'Dim0Min'))
                x = np.arange(npoints) / (npoints-1)  * Dim0Range + Dim0Min
                
                Dim1Range = float(config.get(datatag,'Dim1Range'))
                Dim1Min = float(config.get(datatag,'Dim1Min'))
                y = np.arange(nlines) / (nlines-1) * Dim1Range + Dim1Min
                
                Dim2Range = float(config.get(datatag,'Dim2Range'))
                Dim2Min = float(config.get(datatag,'Dim2Min'))
                img = (img + 2**(nbits-1)) / (2**nbits-1) * Dim2Range + Dim2Min
                # z_value = (z_data + 2^(SaveBits-1)) / (2^SaveBits-1)  * Dim2Range + Dim2Min
                imgdic[cgtag] = [img,x,y]
                i=i+1
                units = [config.get(datatag,'Dim0Unit'),
                         config.get(datatag,'Dim1Unit'),
                         config.get(datatag,'Dim2Unit')]
                
                #if units != ['m','m','m']: 
                    #raise NotImplementedError('unknown units in read_nid: ',units)
                print(units)
                
                #print(points)
                #ReadBinData()
                logging.info(cgtag+' read')
                #print (cgtag+' read')
            except NoOptionError:
                logging.info('option '+cgtag+' not found')
                #print('option '+cgtag+' not found')
                pass
            
    return imgdic

def test_read_nid(file_name=None):
    from pySurf.data2D_class import Data2D
    from matplotlib import pyplot as plt
    if file_name is None:
        file_name=r'..\test\input_data\AFM\02_test.nid'
    
    datadic = read_nid(file_name)
    data,x,y = datadic['Gr0-Ch1']
    print("read data of shape",data.shape)
    
    d = Data2D(data,x,y,units=['mm','mm','um'],scale=[1000.,1000.,1000000.]) 

    d.plot()
    plt.show()
    return d

if __name__ == "__main__":
    
    datafolder = r'G:\My Drive\progetti\c_overcoating\esperimenti\20210129_dopamine\20210224_dopamine_clean'
    fn = 'Image00053.nid'
    file_name =  os.path.join(datafolder,fn)

    d = test_read_nid(file_name)



# -*- coding: utf-8 -*-
"""
It is a library of functions for manipulation or related to optics, acting on a couple of lists x and y, of same length, representing a profile. 

It has only a simple reader (a thin wrapper around `np.genfromtxt`) and a function `register_profile` to align and rescale x and y. Since all functions act on single vectors x and y, more complex readers are not needed here and are left to profile_class. 

Created on Sun Mar 06 16:06:48 2016

@author: Vincenzo Cotroneo
@email: vincenzo.cotroneo@inaf.it
"""
from dataIO.span import span
import os
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
import pdb
from dataIO.functions import update_docstring
from scipy.stats import binned_statistic
## profile creation

def line(x,y=None):
    """return line through end points of x,y.
	x and y are vectors, can be of different length (e.g. y can be a 2-elements vector).
    If only one vector is provided, it is assumend as equally spaced points.
"""
    x=np.array(x)
    if y is None:
        y=x
        x=np.arange(len(y))
    y=np.array(y)
    
    #account for nan
    sel=~np.isnan(y)
    if sel.any():
        y0=[y[sel][0],y[sel][-1]]
        if len(x)==len(y):
            x0=x[sel]
        else:
            x0=x
        L=span(x0,size=1)
        return (x-x0[0])*(y0[-1]-y0[0])/L+y0[0]
    else:
        return y
        
def make_signal(amp,x=None,L=None,N=None,nwaves=None,phase=0,ystartend=(0,0),noise=0.,minus_one=False):
    """
    Build a signal of length L and number of points N, as a sum of a cosinusoid, a line and a noise. minus_one remove last point (just a convenience e.g. for a periodic profile), note that in 
    this case the returned x corresponds to values returned by
    np.arange e .linspace, however it needs to be called with N+1
    points, so the last one can be excluded and N points returned,
    keeping intervals consistent (this might change in future versions).
    Signal is generated on `x` if this is provided.
    Otherwise it is generated on length `L` (can be range).
    Phase adds a phase in radians (the armonic component of signal is defined as `amp*np.sin(2*np.pi*x/L*nwaves+phase)`).
    
    VC 2020/09/11 added option for L as range and x on which to generate signal.
    2020/07/17 make signal a cosine (it was a sine), because more consistent with real part of imaginary number.
    VC 2020/06/27 horrible interface with args, replace with kwargs with defaults.
    Added phase.
    OLD CODE NEEDS TO be UPDATED! TODO: file search
    """
    
    #pdb.set_trace()
    
    # L is the length (starting from zero) or the interval,
    #    on which the signal is generated (phase and nwaves are applied wrt this interval). `x` are the datapoints on which the signal is calculated (defaults to L).
    
    #Note that L and nwaves (togheter with phase) determine the design parameters of the cosinusoid, while x and N the coordinates on which to calculate values.
    
    if x is None:
        if N is None:
            raise TypeError ("N is needed if x is not provided.")
        elif L is None: #N but not L
            L = [0, N-1]      
    else:
        if N is not None:
            raise TypeError ("N is not needed if x is provided.")            
        if L is None:
            L = span(x)        
    
    #after this, if L was not None, it is set to (2-el) interval start and end for the signal generation.
    #pdb.set_trace()
    
    if np.size(L) == 1:
        assert L is not None  #E' stata aggiustata in base a x e a N.
        L = [0,np.ravel(L)[0]] #this is to get scalar element in every case L =[5], L =[[5]], etc.. not sure it is needed and if this is a good way of doing it.
    elif np.size(L) != 2:
        raise ValueError ("Invalid size for L")

    # generate x cocrdinates if not provided
    if x is None:
        x = np.linspace(L[0],L[1],N,dtype=float) #set to N integers
        #x = np.linspace(0,L,N,dtype=float) #set to N integers
        #x=np.arange(N,dtype=float)/(N-1)*L #set to N integers
#x=L[0] + np.arange(N,dtype=float)/(N-1)*(L[1]-L[0]) 
        
    l=line(x,ystartend)
    if nwaves is not None:
        y=amp*np.cos(2*np.pi*(x-L[0])/span(L,size=True)*nwaves+phase) 
    y = y + noise*np.random.random(N)+l  #apply noise and line at once.
    if minus_one:
        x,y=x[:-1],y[:-1]
    return x,y

def test_make_Signal():# test make_signal
    
    x = np.linspace(-0.5,0.5,100)
    plt.clf()

    print("X span: %s N: %i"%(span(x),len(x)))
    #plt.plot(*make_signal(2.,x,nwaves=3))

    #plt.plot(*make_signal(1.,x,nwaves=3,phase=np.pi/2),'o')
    plt.plot(*make_signal(1.,nwaves=3,L=10,N=100),'o')
    plt.plot(*make_signal(2.,nwaves=3,L=[1,11],N=100),'x')
    plt.plot(*make_signal(2.,nwaves=3,L=[1,11],N=100,phase=np.pi/2))
    plt.plot(*make_signal(2.,x,nwaves=3,L=100,phase=np.pi/2))
    #plt.plot(*make_signal(2.,nwaves=3,N=100,phase=np.pi/2))

def make_circle(x,c,r,sign=1):
    """plot positive part if sign is positive, negative if negative."""
	
    y=np.sqrt(R**2-(x-c[0])**2)+c[1]
    return x,y*np.sign(sign)
    
def find_internal_interval(x,y):
    """find the largest interval of profile with all valid point.
    See pySurf.data2D.find_internal_rectangle."""
    raise NotImplementedError
    #return x,y
    
def remove_nan_ends(x,y=None,internal=False):
    """remove all y points at the two ends of profile that contains only nans in y. If internal is set, return the
    internal crop (largest internal segments without nans)."""
    
    if y is None:
        y = x
        x = np.arange(np.size(y))
    
    #resuls are obtained by slicing,
    nanpts=np.where(~np.isnan(y))[0] #indices of columns containing all nan
    if np.size(nanpts)>0:  
    
        try:
            istart=nanpts[nanpts==np.arange(len(nanpts))][-1]+1
        except IndexError:
            istart=0
            
        try:
            istop=nanpts[np.arange(data.shape[1])[-len(nanpts):]==nanpts][0]
        except IndexError:
            istop=np.size(y)+1
        x=x[istart:istop]
        y=y[istart:istop]

    if internal:
        r=find_internal_interval(x,y)
        x,y=crop_profile(x,y,*r)

    return x,y
    
#profile fitting

def polyfit_profile (x,y=None,degree=1,P=np.polynomial.Legendre):
    """Use one of the poynomial bases in 
   (default to Legendre), and return polynomial of given `degree` that fits the input profile `x`.
    
    Temporary function used to separate the numpy layer from the profile, as preferred implementation switched from np.polynomial functions to classes in np.polynomial.Polynomial, which is more flexible, but also has a more complex descriptor (*).
    This function is thinly wrapped by level_profile, the two will eventually merge, maybe with a flag `fit` to return fit or residuals. 
    
    (*) Note that np coefficients are not directly the polynomial coefficient, e.g. [2,4] doesn't necessarily mean 2 + 4 * x 
    There are two poorly documented parameters `window` and `domain` (both default to -1,1) which regulate the scaling of polynomial result. This means that the polynomial coefficients are first calculated on `window` (default [-1,1]) and the result returned shifted to match `domain` (equal to range of x).
    
    Most common usage is to set window to an interval on which the polynomial
    base is well behaving, and domain to the range of x.
    
    see: https://stackoverflow.com/questions/52339907/numpy-polynomial-generation/52469490
    https://github.com/numpy/numpy/issues/9533
    A detailed example of why using window and domain is here:
    https://numpy.org/doc/stable/reference/routines.polynomials.classes.html
    
    Some of the Polynomials bases are well-behaving, or orthonormal only on a given `window` of x (e.g. Legendre on [-1,1]), we usually want

In using Chebyshev polynomials for fitting we want to use the region where x is between -1 and 1 and that is what the window specifies. However, it is unlikely that the data to be fit has all its data points in that interval, so we use domain to specify the interval where the data points lie. When the fit is done, the domain is first mapped to the window by a linear transformation and the usual least squares fit is done using the mapped data points. The window and domain of the fit are part of the returned series and are automatically used when computing values, derivatives, and such. If they aren’t specified in the call the fitting routine will use the default window and the smallest domain that holds all the data points.

(that seems in contradiction with numpy documentation that states both default to one).    
    
"""    

    if y is None:
        y=x
        x=np.arange(len(y))
    
    goodind=~np.isnan(y) #isolate valid data
    if goodind.any():
        y0=y[goodind]
        x0=x[goodind]
    else:
        x0,y0 = x,y    
    
    res = P.fit(x0,y0,degree) #fit only valid data
    #coeff=np.polyfit(x0,y0,degree)
    
    result = res(x) #result = np.polyval(coeff,x)
    result[~goodind] = np.nan #set back nans if any
    
    return result
 
        
##PROFILE I/O
    
def register_profile(x,y,scale=(1,1),
    strip=False,crop=None,center=None,*args,**kwargs):
    """get x,y and register them using usual set of parameters.

    registering operation are performed in the following order and are:
        scale: scale the two axis of the `scale` factor, if sign is changed, reorder.
        strip: if True, strip all nans at the border of the data.
        crop: list of ranges (2 axis) to pass to profile.crop_profile
        center: the final position of profile center, if 1 element, only x is centered, if 2 elements, y is centered also.
    This means e.g. that data are first cropped (or cleaned of invalid data) than centered. This means that the value puts in the provided coordinate(s) the center of points after cropping operations.

    unexpected parameters passed to register_data are ignored (*args and **kwargs are not used, just suppress error).

    `load_profile` doesn't automatically call `register_data`,
        this might be different from pySurf.
    """

    x=x*scale[0]
    y=y*scale[1]
    
    #if x is inverted, invert data and orient as cartesian.
    #this maybe can be move somewhere else, maybe in class set_data,x,y
    # but be careful to avoid double reflection.
    
    if x[-1]<x[0]:
        x=x[::-1]
        y=y[::-1]  #added 2021/06/15 it was never tested?
        #data=np.fliplr(data)  #removed, no data here, from data2D
        #x=x-min(x)    # why?

    #adjust crop and scales
    if strip:  #messes up x and y
        #print("WARNING: strip nans is not properly implemented and messes up X and Y.")
        x,y = remove_nan_ends(x,y)

    #doing cropping here has the effect of cropping on cartesian orientation,
    #coordinates for crop are independent on center and dependent on scale and x bin (pixel size).
    #center is doing later, resulting in center of cropped profile only is placed in the suitable poition.
    if crop is None:
        crop=[None,None]
    x,y=crop_profile(x,y,*crop)

    #center data if center is None, leave unchanged
    if center is not None:
        assert len(center)>=1
        x=x-(np.max(x)+np.min(x))/2.+center[0]
        if len(center)==2:
            y=y-(np.nanmax(y)+np.nanmin(y))/2.+center[1]

    return x,y

def load_profile(file,*args,**kwargs):
    """The simplest file loader using np.genfromtxt.
        Returns x and y."""
        
    return np.genfromtxt(file,unpack=True,*args,**kwargs)

def save_profile(filename,x,y,**kwargs):
    """Use np.savetxt to save x and y to file """
    np.savetxt(filename,np.vstack([x,y]).T,**kwargs)
        
#PROFILE OPERATIONS        

def get_stats(x,y=None,units=None):
    
    """"""
    if units is None:
        u = ["",""]
    elif np.size(units) == 1:
        u=[units,units]
    else:
        u=units
    assert np.size(u) == 2
    
    stats=['RMS:%3.3g %s'%(np.nanstd(y),u[1]),'PV_X:%3.3g %s, PV_Y:%3.3g %s'%(span(x,size=1),u[0],span(y,size=1),u[1])]
    return stats
    
def level_profile (x,y=None,degree=1):
    """return profile after removal of a (Legendre) polynomial component.
    polyfit_profile has an option to change the polynomial base if Legendre is not the desired base.
"""
    if y is None:
        y=x
        x=np.arange(len(y))
    
    return x,y-polyfit_profile(x,y,degree) 

def sort_profile (x,y,reverse=False):
    """sort profile on the base of x. """

    i = x.argsort()    
    if reverse:
        i=i[::-1]
    return x[i],y[i]       

def remove_profile_outliers(y,nsigma=3,includenan=True):
    """remove outliers from a profile by interpolation.
    y is modified in place. Mask of outliers is returned. """
    
    mask=np.abs(y-np.nanmean(y))<(nsigma*np.nanstd(y)) #points to keep
    if includenan:
        mask[np.isnan(y)]=False
        
    if mask.any()==False:
        return y*np.nan  #all invalid values, transform to nan so it can be
            #detected on call. numeric values cannot be all out of nsigma, unless small.

    if not mask.all():
        x=np.arange(len(y))
        y[mask==False] = np.interp(x[mask==False], x[mask], y[mask])
        
    return mask==False
	

    
"""
def movingaverage(values,window,mode='full'):
    weigths = np.repeat(1.0, window)/window
    #including valid will REQUIRE there to be enough datapoints.
    #for example, if you take out valid, it will start @ point one,
    #not having any prior points, so itll be 1+0+0 = 1 /3 = .3333
    smas = np.convolve(values, weigths, mode)
    return smas # as a numpy array
"""   

def movingaverage(values,window,method='same',*args,**kwargs):
    """ Convolve vector ``values`` with a ``window`` using ``np.convolve``.
    
    Window can be provided as vector or as integer (number of points, must be odd,
    or it is rounded to lower odd number. (see np.convolve and np.scipy.convolve).
    
    2022///06/28 added option to use smoothing window."""
    
    try:
        if len(window) == 1:
            window = window[0]
            assert int(window/2.)*2==(window-1)
            weights = np.repeat(1.0, window)/window
        else:
            pass
    except TypeError:
        assert int(window/2.)*2==(window-1)
        weights = np.repeat(1.0, window)/window
    
    print(values,weights)
    #including valid will REQUIRE there be enough datapoints.
    #for example, if you take out valid, it will start @ point one,
    #not having any prior points, so itll be 1+0+0 = 1 /3 = .3333
    smas = np.convolve(values, weights, method,*args,**kwargs)
    for i in np.arange(int(window/2.)+1):
        smas[i]=np.mean(values[:2*i+1])
        smas[-i]=np.mean(values[-2*i-1:])
    return smas # as a numpy array

def rebin_profile(x,y,*args,**kwargs):
    """ Flexible rebin of a profile. Reduces number of points without losing information.
    Uses `stats.binned_statistics` of which keeps the interface. """
    ss=stats.binned_statistic(x,y,statistic='mean',*args,**kwargs)
    x2=np.array([(x+y)/2. for x,y in zip(ss[1][:-1],ss[1][1:])])  #centra su punto centrale
    y2=ss[0]
    return x2,y2  
    rebin_profile = update_docstring(rebin_profile, stats.binned_statistics)

def crop_profile(x,y=None,xrange=None,yrange=None,*args,**kwargs):
    """Crop a profile to range (can be set to None) in x and y.
    Experimental, crop on y leaves holes in profile."""
    #qui sarebbe utile impostare gli argomenti come x e y necessariamente args e i ranges kwargs.
    #Si potrebbe anche implementare una funzione piu' potente come clip in IDL
    
    if y is None:
        y=x
        x=np.arange(np.size(y))
    
    if xrange is None: 
        xrange=[None,None]
    if xrange[0] is None:
        xrange[0]=np.nanmin(x)
    if xrange[1] is None:
        xrange[1]=np.nanmax(x)

    sel=(x>=xrange[0])&(x<=xrange[1])
    x=x[sel]
    y=y[sel]
    
    if yrange is None: 
        yrange=[None,None]
    if yrange[0] is None:
        yrange[0]=np.nanmin(y)
    if yrange[1] is None:
        yrange[1]=np.nanmax(y)

    sel=(y>=yrange[0])&(y<=yrange[1]) 
    

    return x[sel],y[sel]  

def resample_profile(x1,y1,x2,y2=None):
    """resample y1 (defined on x1) on x2.
    Both x1 and y1 need to be set for input data,
        x2 is returned together with interpolated values as a tuple.
    y2 is not used and put for consistency (can be omitted).
    
    """
    if np.any(np.diff(x1) <= 0): 
        if np.any(np.diff(x1) >= 0): 
            raise ValueError ('x1 must be monotonic for interpolation')
    if np.any(np.diff(x2) <= 0): 
        if np.any(np.diff(x1) >= 0): 
            raise ValueError ('x2 must be monotonic for interpolation')    
    
    y2 = np.interp(x2,x1,y1)
     
    return x2,y2

def sum_profiles(x1,y1,x2,y2,*args,**kwargs):
    return x1,y1+resample_profile(x2,y2,x1)[1]  
        
def subtract_profiles(x1,y1,x2,y2,*args,**kwargs):
    return x1,y1-resample_profile(x2,y2,x1)[1] 

def merge_profile(x1,y1,x2,y2):
    """stitch profiles"""
    raise NotImplementedError("see example of psd merging in G:\My Drive\progetti\read_nid.ipynb")
    
    
def merge_profiles(profiles,ranges=None):
                    
#def merge_profiles(ranges, profiles, labels,xrange=None,yrange=None,
#                    outname=None):
    """ da trim_psds_group in ICSO2020_review
    
        Makes a single profile trimming and averaging multiple ones 
        according to selection.
    
        this is lauched to trim a single set of psds related to same sample (or same context).
        ranges, profiles, labels are lists with same number of elements,
        describing respectively: 
        ranges for each psd (if None, the psd is not included in the output).
            If set to None, include all data.
        labels to be used in plot.
        outname: if provided generate plot of trim and txt with resulting psd.
        
        """
    if ranges is None:
        ranges = np.repeat([[None],[None]],
                           len(profiles),axis=1).T
    xtot,ytot,bins,xvals,yvals = [] , [], [], [], []
    
    #plt.figure(figsize=(12,6))
    #for ran,pr,lab in zip(ranges,profiles,labels):
    for ran,pr in zip(ranges,profiles):
        if ran is not None:
            #print(d)
            x,y = pr
            if x[0] == 0:
                x=x[1:]
                y=y[1:]
            xx,yy = crop_profile(x,y,ran)
            #plot_psd(xx,yy,label=lab,units=['um','um','nm'])
            xtot.append(xx)
            ytot.append(yy)
        """
        plt.legend( prop={'size': 12},loc = 1) #bbox_to_anchor=(1.05, 1), loc='upper left')
        plt.xlim(xrange)
        plt.ylim(yrange)
        if outname:
            plt.title(os.path.basename(outname))
        plt.grid( which='minor', linestyle='--')
        plt.tight_layout()
    if outname:
        plt.savefig(fn_add_subfix(outname,'_trimpsds','.png'))
    """
    
    ### The following part is of interest only for rebinning,
    ###   not included in this version's output
    ###   because result will in general not be smooth in transitions between
    ###   different intervals.
    ###   if intervals are not overlapping, it is irrelevant.
    # makes rebinning and averaging.
    # we want to start from lower x (assuming inputs are passed not in order),
    # keep x in non overlapping regions, while averaging on common regions.
    # psd frequencies (x) are not equally spacing. If I do average, I get
    #   spikes when typically lower freqs of each psd are more spaced than higher freq
    #   of same psds. Then overlapping intervals of two psds typically have one with
    #   broader spacing. Most of intervals have points only from psds with tighter spacing
    #      when points from the other enter, you get spike. This is why interpolation is needed.
    #sort groups in xtot in ascending order
    pmin=[a.min() for a in xtot]
    igroup =np.argsort(pmin)
    for i in igroup:
        x = xtot[i]
        y = ytot[i]
        if len(bins) == 0:
            bins.append(x)
            xvals.append(x)
            yvals.append(y)
        else:
            sel = x>max(bins[-1])
            xint = np.hstack([ bins[-1][bins[-1]>=min(x)] , x[sel] ])
            #pdb.set_trace()
            xvals.append(xint)
            yvals.append(np.interp(xint,x,y))
            if any(sel):
                #pdb.set_trace()
                #resample second vector on common range
                bins.append(x[sel])
    xtot=np.hstack(xtot)
    ytot=np.hstack(ytot)
    xvals=np.hstack(xvals)
    yvals=np.hstack(yvals)
    xbins = np.hstack(bins)
    ybins = binned_statistic(xvals,yvals,bins=xbins,statistic='mean') [0]
    ###
    
    """
    plot_psd(xbins[:-1],ybins,label='binned',units=['um','um','nm'],
             linestyle='--')
    #plot_psd(xtot,ytot,label='total',units=['um','um','nm'])
    if outname:
        save_profile(fn_add_subfix(outname,'_binpsd','.dat'),xbins[:-1],ybins)
    
    if outname:
        plt.savefig(fn_add_subfix(outname,'_trimpsds','.png'))
    """
    
    return xtot, ytot    

def calculate_barycenter(x,y):
    """return the x of barycenter using y as weight function."""
    return np.sum(x*y)/len(x)
    
def reflect_profile(xx,yy,center=None,split=False,xout=None):
    """Return the profile x, y mirrored around x=center.
    If split is set return a list of the two profiles >,< 
    as [[xpos,ypos],[xneg,yneg]].
    If xout is passed it is used to interpolate the double
    profiles. If split is not selected and the initial profile
    is not symmetrical can give unexpected results. 
    In case split is selected and center exactly matches one of the input points,
    center point is added to both profiles.
    """
    x=xx.copy()
    y=yy.copy()
    
    if center is None:
        center=calculate_barycenter(x,y)
    
    mpos = (x >= center) #mask
    x[~mpos]=2*center-x[~mpos]
    
    if split:
        if center in x:
            mneg=np.logical_not(mpos)
            mneg[x==center]=True
        x0,y0=[x[mneg],y[mneg]],[x[mpos],y[mpos]] #note that x0 and y0 are no more x and y, rather two profiles returned (no sorting)
        if xout is not None:
            x0,y0=[[xout,np.interp(xout,np.sort(xx),yy[xx.argsort()])] for xx,yy in [x0,y0]]
            #x0,y0=[[xout,np.interp(xout,xx,yy)] for xx,yy in [x0,y0]]
    else:
        i0=x.argsort()
        x0=x[i0]
        y0=y[i0]        
        if xout is not None:
            y0=np.interp(xout,x0,y0)
            x0=xout
    
    return x0,y0
        
        
    
def calculate_HEW(x,y,center=None,fraction=0.5,profile=False):
    """calculate HEW around center from profile x and y by integrating the two sides. If center is None, barycenter is calculated.
    Radius is returned that gives integrated height equal to `fraction`
    of total integral.
    If `profile` is set, a profile x0,y0 of the integrated intensity is returned, where x0 is distance from `center` and y0 is integrated y."""
    
    x2 = np.array(x)
    y2 = np.array(y)
    
    if center is None:
        center=calculate_barycenter(x,y)
    
    xe,ye=reflect_profile(x,y,center=center)
    intprof=[np.trapz(ye[:i],xe[:i]) for i in np.arange(len(xe))]
    he=intprof[-1]*fraction
    if profile:
        return xe,np.array(intprof)
    else:
        return xe[np.count_nonzero(intprof<=he)]
    
    
def plot_HEW(xout,yout,center=None,fraction=0.5):
    """Plots details of HEW calculation."""
    
    cs=' barycenter' if center is None else  'x=%.3g'%center
    #plt.title('HEW centered about'+cs)

    xi,yi=calculate_HEW(xout,yout,profile=True,center=center)
    plt.xlabel('angular distance from barycenter (arcsec)')
    plt.ylabel('Intensity (a.u.)', color='b')

    plt.plot(*reflect_profile(xout*206265.,yout,center=center),
             label='Radial intensity')
    #ax1=plt.gca()
    ax2 = plt.gca().twinx()
    plt.plot(xi*206265.,yi/np.nanmax(yi),'r',label='Normalized integral')

    ax2.set_ylabel('Integrated PSF, normalized')
    #ax2.tick_params('y', colors='r')

    plt.legend(loc=7)
    
    hew=calculate_HEW(xout,yout,center=center)
    plt.axhline(yi[np.count_nonzero(yi<np.nanmax(yi)*fraction)]/np.nanmax(yi),
                ls='--',c='y')
    plt.axvline(hew*206265.,ls='--',c='y')
    
    plt.show()   
    return hew

#PROFILE STATS AND DERIVED QUANTITIES
def PSF_spizzichino(x,y,alpha=0,xout=None,energy=1.,level=True, HEW=True):
    """Try to use spizzichino theory as in PR notes to calculate Hthe PSF,
    return a vector of same length as xout.
    alpha is incidence angle from normal in degrees, alpha= 90 - shell slope for tilt removed profiles.
        Tilt can be included in the profile, in that case alpha is left to 0
       (total slope must be <0 and alpha>0, this means that profile with tilt is expected to be <0).
    xout can set the output intervals in theta on the focal plane (from specular angle), if not set 512 points are used.
    Lambda is wavelength in keV."""
    
    lambda_mm=12.398425/energy/10**7
    
    if xout is None:
        lout=1001
    else:
        lout=len(xout)
    L=span(x,size=True)
    deltax=L/(len(x))
    
    #calculate and remove slope as alpha. Profile tilt removed is yl
    if level: 
        slope=line(x,y)
        yl=y-slope
        #adjust incidence angle to include the slope removed from profile leveling. 
        # Increasing profile is positive slope angle:
        alpha=alpha*np.pi/180-np.arctan2(y[-1]-y[0],x[-1]-x[0])
    else:
        yl=y
        
    if alpha<=0: raise ValueError
    
    thmax= lambda_mm/(2*deltax*(np.pi/2-alpha))
    #xout is the array of theta for the output
    if xout is None:
        xout=np.linspace(alpha-thmax,alpha+thmax,lout)
    else:
        xout=xout+alpha
        
    scale=np.sqrt(2.)
    I=np.array([np.abs((deltax/L*(np.exp(2*np.pi*1.j/lambda_mm*(x*(np.sin(alpha)-np.sin(theta))-scale*yl*(np.cos(alpha)+np.cos(theta)))))).sum())**2 for theta in xout])
    """
    The above is equivalent to (iterate on all theta in xout):
    I[theta]=np.abs((deltax/L*(np.exp(2*np.pi*1.j/lambda_mm*(x*(np.sin(alpha)-np.sin(theta))-scale*yl*(np.cos(alpha)+np.cos(theta)))))).sum())**2
    """
    
    #if HEW:
    #    calculate_HEW(xout-alpha,I,center=None,fraction=0.5)
        
    return xout-alpha,I

def PSF_raimondiSR(x,y,alpha=0,xout=None,energy=1.):
    """Try to use theory from Raimondi and Spiga A&A2015 to calculate the PSF for single reflection,
    return a vector of same length as xout.
    alpha is incidence angle from normal in degrees, alpha= 90 - shell slope for tilt removed profiles.
    Tilt can be included in the profile, in that case alpha is left to 0
       (total slope must be <0 and alpha>0, this means that profile with tilt is expected to be <0).
    xout can set the output intervals in theta on the focal plane (from specular angle), if not set 512 points are used.
    Lambda is wavelength in keV.
    """
    
    """
    R0 e' necessario ma solo per normalizzare la PSF.  In realtaà se nella formula all'ultima riga sostituisci dr1 = L*sin(alpha) = L*R0/2f (usi la singola riflessione, giusto?) vedrai che R0 se ne va e non serve saperlo. f a questo punto sara' semplicemente la distanza alla quale si valuta il campo, non necessariamente la focale.
    
    dr1 = L*sin(alpha) = L*R0/2f
    
    """
    
    lambda_mm=12.398425/energy/10**7
    
    if xout is None:
        lout=1001
    else:
        lout=len(xout)
    L=span(x,size=True)
    deltax=L/(len(x))
    
    #calculate and remove slope as alpha. Profile after tilt removal is yl
    slope=line(x,y)
    yl=y-slope
    #adjust incidence angle to include the slope removed frorm profile leveling. 
    # Increasing profile is positive slope angle:
    alpha=alpha*np.pi/180-np.arctan2(y[-1]-y[0],x[-1]-x[0]) 
    if alpha<=0: raise ValueError
    
    thmax= lambda_mm/(2*deltax*(np.pi/2-alpha))
    #xout is the array of theta for the output
    if xout is None:
        xout=np.linspace(alpha-thmax,alpha+thmax,lout)
    else:
        xout=xout+alpha
    
    dR1=L*np.sin(alpha)  #L1
    R0=1  #F*np.tan(4*alpha)
    d20=1  #randdomly set things to 1
    z1=1
    
    
    
    PSF=1/2*F/(L*lambda_mm)*np.abs([np.sqrt(y/d20)*np.exp(-2*np.pi*1.j/lambda_mm*(d20-z1+x**2/(2*(S-x))))].sum()*deltax)**2  #L1
    
    
    """ spizzichino original:
    scale=np.sqrt(2.)
    I=[np.abs((deltax/L*(np.exp(2*np.pi*1.j/lambda_mm*(x*(np.sin(alpha)-np.sin(theta))-scale*yl*(np.cos(alpha)+np.cos(theta)))))).sum())**2 for theta in xout]
    """
    
    """
    The above is equivalent to (iterate on all theta in xout):
    I[theta]=np.abs((deltax/L*(np.exp(2*np.pi*1.j/lambda_mm*(x*(np.sin(alpha)-np.sin(theta))-scale*yl*(np.cos(alpha)+np.cos(theta)))))).sum())**2
    """
    return xout-alpha,I    
 
'''
#P.S. I suspect this is the same as removing average tilt. Proof??
def autotilt(x,y):
    """Transform a profile by tilting in a way that two contact points are on same horizontal line, return removed line."""
    if delta[0]==0 and delta[-1]==0
    L=span(x,size=1)
    line=x*(delta[-1]+delta[0])/L-delta[0]
    delta=delta-line  #start from endpoints
    i,j=delta.argsort()[:2]   #index of two maxima
'''

##TESTS AND USE

def test_reflect(xx=None,yy=None,center=None,outfolder=None):
    
    print ("specularly reflect a profile about a center position on x axis")
    xx=np.arange(30)
    yy=-0.03*xx**2+0.2*xx-5
    center=15
    
    
    plt.close('all')
    plt.figure()
    plt.plot(xx,yy,label='starting profile')
    plt.plot(*reflect_profile(xx,yy,center=center),label='reflect about 15')
    plt.plot(*reflect_profile(xx,yy,center=center),'o')
    for i,p in enumerate(reflect_profile(xx,yy,center=center,split=1)):
        plt.plot(*p,'-.',label='reflect about 15, split#%i'%i) 
    plt.legend(loc=0)

    plt.figure()
    xout=15+np.arange(5)*2
    plt.title('test interpolated output')
    plt.plot(xx,yy,label='starting profile')
    plt.plot(*reflect_profile(xx,yy,center=center,xout=xout),label='reflect about 15')
    plt.plot(*reflect_profile(xx,yy,center=center,xout=xout),'o')
    for i,p in enumerate(reflect_profile(xx,yy,center=center,split=1,xout=xout)):
        plt.plot(*p,'-.',label='reflect about 15, split#%i'%i) 
    plt.legend(loc=0)
    
    plt.show()
    print('done!')
    return xx,yy

def test_HEW():
    #datafile=r'test\01_mandrel3_xscan_20140706.txt'
    print ("uses Spizzichino's formula to predict PSF on sinusoidal ")
    x=np.linspace(0,300.,100)
    y=np.cos(6*np.pi*x/span(x,size=True))/60000.
    xout=np.linspace(-5,5,1000)/206265.
    #x,y=np.genfromtxt(datafile,unpack=True,delimiter=',')
    #y=y-line(x,y)
    plt.figure('profile')
    plt.clf()
    plt.title('profile')
    plt.plot(x,y)
    plt.plot(x,line(x,y))
    plt.xlabel('Axial Position (mm)')
    plt.ylabel('Profile height (mm)')
    plt.figure('PSF')
    plt.clf()
    alpha=89.79
    plt.title('PSF, alpha=%f6.2'%alpha)
    xout,yout=PSF_spizzichino(x,y,alpha=alpha,xout=xout)
    plt.xlabel('angular position around alpha (arcsec)')
    plt.ylabel('Intensity (a.u.)')
    plt.plot(xout*206265.,yout)
    plt.show()
    print('done!')
    return xout,yout
    
if __name__=="__main__":
    test_HEW()
    


# -*- coding: utf-8 -*-

"""  This module contains functions able to read raw data from different formats. With default arguments returns data,x,y; return header if `header` is set to True. 
    Provides common interface to several instrument reader
    routines. Functions are made to be imported in intrumentReader (from which functions were originally copied) and not to be
    directly called.

2020/05/25 Moved here csv4D_reader

2020/03/18 Transfer here functions from _instrument_reader.

2018/09/26 note that these routines shouldn't take *args,**kwargs arguments, unless they need to pass it to another
    function or want to be forgiving of wrong argument passed to them.
    
 Returns data and header from file, with minimal data
     
2018/09/26 New refactoring of data reader. Vincenzo Cotroneo vcotroneo@cfa.harvard.edu"""

import numpy as np
from matplotlib import pyplot as plt
import os
from astropy.io import fits

from pySurf.points import get_points
from pySurf.points import crop_points
from pySurf.points import points_find_grid
from pySurf.points import resample_grid
from pySurf.readers.read_sur_files import readsur
from pySurf.readers.read_metropro_files import readMetroProData
#from utilities.imaging.man import stripnans
from pySurf.data2D import data_from_txt

from dataIO.read_pars_from_namelist import read_pars_from_namelist
from dataIO.fn_add_subfix import fn_add_subfix
from IPython.display import display

from .nid_reader import make_channel_tags, read_datablock

import pdb


def points_reader(wfile,header=False,*args,**kwargs):
    """Read a processed points file in format x,y,z as csv output of analysis routines."""
    if header: return None
    w0=get_points(wfile,*args,**kwargs)
    w=w0.copy()
    x,y=points_find_grid(w,'grid')[1]
    pdata=resample_grid(w,matrix=True)
    return pdata,x,y

def datzygo_reader(wfile,header=False,*args,**kwargs):
    """read .dat binary files (MetroPro/Zygo)."""
    
    '''from manual: Complete maps of the header formats can be obtained by running the dat_test.exe 
program that is included with MetroPro in folder r"C:\MetroPro\Bin". Type the following at 
a command prompt: 
dat_test  –maps  >  maps.txt 
File maps.txt will contain the output. '''

    d1,head,d3,d4=readMetroProData(wfile,*args,**kwargs)
    #pdb.set_trace()
    if header or kwargs.get('header',False): return head
    try:
        x0=head['X0']
    except KeyError:
        x0=0
    try:
        y0=head['Y0']
    except KeyError:
        y0=0
    try:
        xyscale=head['cameraRes']
    except KeyError:
        xyscale=1.
    # pdb.set_trace()
    #data,x,y=d1,x0+np.arange(np.shape(d1)[1]),y0+np.arange(np.shape(d1)[0])  
    # data, x, y are already in the appropriate units, but xy offsets were not included.
    # units are meters.
    data,x,y=d1,x0*xyscale+d3,y0*xyscale+d4

    return data,x,y


def csv4D_reader(wfile,ypix=None,ytox=None,header=False,delimiter=',',endline=True,skip_header=12,*args,**kwargs):
    """read csv data in 4sight 4D format.
    12 lines header with info in namelist format, uses `xpix`, `aspect` and `wavelength` if available.
    Note that standard csv format ends line with `,` which adds an extra column.
    This is automatically removed if `endline` is set to True.
    Note also that typically instruments invert y axis, but this is not implemented at the moment. 
    TODO: change y scale after code reorganization.
    2020/07/14 read data directly with `np.genfromtxt`,
    rather than uselessely launching the wrapper `data2D.data_from_txt`."""

    head=read_pars_from_namelist(wfile,': ') #this returns a dictionary, order is lost if header is returned.
    if header:
        return '\n'.join([": ".join((k,v)) for (k,v) in head.items()])+'\n'

    if ypix == None:
        try:
            ypix=np.float(head['xpix'])
        except KeyError:
            ypix=1.

    if ytox == None:
        try:
            ytox=np.float(head['aspect'])
        except KeyError:
            ytox=1.

    try:
        zscale=np.float(head['wavelength'])
    except KeyError:
        zscale=1.
    
    #from pySurf.data2D import data_from_txt
    #data=np.genfromtxt(wfile,delimiter=delimiter,skip_header=12)
    #data=data_from_txt(wfile,delimiter=delimiter,skip_header=skip_header)[0]
    data=np.genfromtxt(wfile,skip_header=skip_header,delimiter=delimiter)
    #pdb.set_trace()
    #this defines the position of row/columns, starting from
    # commented 2018/08/28 x and y read directly
    if endline:
        data=data[:,:-1]
    ny,nx=data.shape
    x=np.arange(nx)*ypix*ytox*nx/(nx-1)
    y=np.arange(ny)*ypix*ny/(ny-1)
    data=data*zscale
    #data.header=head


    return data,x,y


def csvZygo_reader(wfile,intensity=False,header=False,xyz=False,*args,**kwargs):

    """read .csv zygo files (and .xyz).

    Height map is returned unless intensity is set to True to return Intensity map in same format.
    Metadata are reaad from header, some can be overridden by
      arguments in kwargs:
      
      ypix defaults to CameraRes
      ytox defaults to 1. 
      zscale defaults to WavelengthIn*1000000. #original unit is m, convert to um
    
    ### XYZ format:
    XYZ Data File Connected Phase Data
    The data in this section is organized by phase origin. Each line contains three pieces of
    data. The first two columns contain the column (y) and row (x) location of the data,
    beginning at the phase origin. The third number on the line can be either the character
    string "No Data" or a floating-point number corresponding to the measurement in
    microns. To convert these measurements to ‘zygos’, use the following formula. (The
    names in parenthesis refer to the Binary Data Format field names)
    The data in the file is in microns. To convert to meters:
    For Low resolution (0) phase data:
    Multiply third column data in by:
     (4096/ (IntfScaleFactor*ObliquityFactor*WavelengthIn*1000000))
    For High resolution (1) phase data:
    Multiply third column data by:
     (32768/ (IntfScaleFactor*ObliquityFactor*WavelengthIn*1000000))
    The PhaseRes is the first value of the eleventh line of the header. Convert the result to an
    integer to be stored in the binary file. This will cause round-off error, but amounts to a
    less than one Angstrom variance from the original data. 

    ### ASCII data format:
    This section describes the format of a MetroPro ASCII data file. The file is made up of
    three parts: header, intensity data, and phase data. Each part is followed by a line
    containing a sharp (#) character. At least one of the data sets must be present. A
    MetroPro ASCII data file is created by using the dat_to_asc conversion utility. 
    
    -----------------
    
    ASCII Data File Intensity Data
    Each data point is an integer. The data is written 10 data points per line in row-major
    order. Acceptable values are from 0 to the value specified in IntensRange. An invalid
    point is indicated by a value >= 65535. A line containing only a sharp character (#) is
    output after the data. The number of intensity data points is:
    IntensWidth * IntensHeight * NBuckets
    ASCII Data File Connected Phase Data
    Each data point is an integer. The data is written 10 data points per line in row-major
    order. Acceptable values are in the range from -2097152 to +2097151. An invalid point
    is indicated by a value >= 2147483640. A line containing only a sharp character (#) is
    output after the data. The number of connected phase data points is:
    PhaseWidth * PhaseHeight
    The phase data points are in internal units representing a scaled number of fringes. To
    convert a value to waves, multiple by (S * O)/R.
    Where: S = IntfScaleFactor, O
    
    """

    with open(wfile) as myfile:
        raw = myfile.readlines()

    head,d=raw[:15],raw[15:]
    if header: return head

    nx,ny = list(map( int , head[8].split()[:2] ))
    # coordinates of the origin of the
    #connected phase data matrix. They refer to positions in the camera coordinate system.
    #The origin of the camera coordinate system (0,0) is located in the upper left corner of the
    #video monitor.
    origin = list(map(int,head[3].split()[:2]))

    #These integers are the width (columns) and height (rows)
    #of the connected phase data matrix. If no phase data is present, these values are zero.
    connected_size=list(map(int,head[3].split()[2:4]))

    #This real number is the interferometric scale factor. It is the number of
    # waves per fringe as specified by the user
    IntfScaleFactor=float(head[7].split()[1])

    #This real number is the wavelength, in meters, at which the interferogram was measured.
    WavelengthIn=float(head[7].split()[2])

    #This integer indicates the sign of the data. The data sign may be normal (0) or
    #inverted (1).
    DataSign = int(head[10].split()[7])
    #This real number is a phase correction factor required when using a
    #Mirau objective on a microscope. A value of 1.0 indicates no correction factor was
    #required.
    Obliquity = float(head[7].split()[4])

    #This integer indicates the resolution of the phase data points. A value of 0
    #indicates normal resolution, with each fringe represented by 4096 counts. A value of 1
    #indicates high resolution, with each fringe represented by 32768 counts.
    phaseres = int(head[10].split()[0])
    if phaseres==0:
        R = 4096
    elif phaseres==1:
        R = 32768
    else:
        raise ValueError

    #This real number is the lateral resolving power of a camera pixel in
    #meters/pixel. A value of 0 means that the value is unknown
    CameraRes = float(head[7].split()[6])
    if CameraRes==0:
        CameraRes=1.

    #pdb.set_trace()
    ypix=kwargs.pop('ypix',CameraRes)
    ytox=kwargs.pop('ytox',1.)
    if xyz:
        zscale=kwargs.pop('zscale',1.) # assume already in um    
    else:
        zscale=kwargs.pop('zscale',WavelengthIn*1000000.)#original unit is m, convert to um

    #pdb.set_trace()
    if xyz:
        tmp = [dd.replace('No Data','nan') for dd in d[:-1]]
        tmp = np.array([l.split() for l in tmp],dtype='float')
        from pySurf.points import resample_grid
        tmp = resample_grid(tmp,matrix=True)
        datasets=[np.array([]),tmp]
        #return datasets[-1]
    else:
        datasets=[np.array(aa.split(),dtype=int)  for aa in ' '.join(map(str.strip,d)).split('#')[:-1]]
    #here rough test to plot things
    d1,d2=datasets  #d1 intensity, d2 phase as 1-d arrays
    
    d1,d2=d1.astype(float),d2.astype(float)
    if np.size(d1) > 0: d1 = d1.reshape(ny,nx)
    if np.size(d2) > 0: d2 = d2.reshape(*connected_size[::-1])
    
    '''    
        datasets=[np.array(aa.split(),dtype=int)  for aa in ' '.join(map(str.strip,d)).split('#')[:-1]]
        #here rough test to plot things
        d1,d2=datasets  #d1 intensity, d2 phase
        d1,d2=d1.astype(float).reshape(ny,nx),d2.astype(float).reshape(*connected_size[::-1])
        d1[d1>65535]=np.nan
        d2[d2>=2147483640]=np.nan
        d2=d2*IntfScaleFactor*Obliquity/R*zscale #in um
    '''
    
    d1[d1>65535]=np.nan
    if not xyz:
        d2[d2>=2147483640]=np.nan
        d2=d2*IntfScaleFactor*Obliquity/R*zscale #in um

    #pdb.set_trace()
    
    dd2=np.zeros([ny,nx])*np.nan #d1 is same size as sensor
    dd2[origin[1]:origin[1]+connected_size[1],origin[0]:origin[0]+connected_size[0]]=d2
    d2=dd2

    #this defines the position of row/columns, starting from
    x=np.arange(nx)*ypix*ytox*nx/(nx-1)
    y=np.arange(ny)*ypix*ny/(ny-1)

    return  (d1,x,y) if intensity else (d2,x,y)

from .read_sur_files import readsur

def sur_reader(wfile,header=False,*args,**kwargs):
    """read .sur binary files."""
    head=readsur(wfile,*args,**kwargs)
    if header: return head

    data,x,y=head.points,head.xAxis,head.yAxis
    del(head.points,head.xAxis,head.yAxis) #remove data after they are extracted from header to save memory.
    return data,x.flatten(),y.flatten()   #are returned as column vector  
        
def read_nid(file_name,index=0,header=False,read_tags=False):
    """read a file nid. Return a list of `data,x,y` for the scans of index in `index`.
    
    `data,x,y` are extracted from files and adjusted by using 
     info from header. A file nid can contains several "scans" 
     (typically different AFM data: phase, amplitude, etc., having different
     units). This function extracts a single scan.
     
    Metadata have a hyerarchical structure, with first level general settings,
    then subfields for singl scans
    if `header` is set, the full file metadata are returned "raw" as list of string,
    it is left to user to extract the subset of metadata corresponding to the 
    extracted data.
    A convenient way to explore the metadata is to convert them to a 
    configparser object. This is already implemented in this routine, where
    values should already include settings and convertions from metadata.
    TODO: consider extracting only relevant metadata (if possible/convenient). 
    
    `read_tags` if set, keys for channels with data are returned (to return also the empty ones call `make_channel_tags`)
    
    Of the old primitive routine `nid_reader.read_nid` and 
    `nid_reader.read_raw_nid`:
    `read_raw_nid` reads `header` as list of strings and `data` as
    single binary block. `read_datablock` is used to extract an image
    from the datablock. `dataIO.config.string_to_config` can be used by converting it to `config`
    object and reading fields."""
    
    from pySurf.readers.nid_reader import read_raw_nid  #,read_nid
    from dataIO.config.make_config import string_to_config
    import logging
     
    
    # all columns of the matrix 
    #ngroups = config.get('DataSet','GroupCount') #number of groups 
    imgdic=[]

    logging.info('reading '+file_name)
    #print('reading '+file_name)
    meta, data = read_raw_nid(file_name)
    if header: return meta
    
    # build a config object `config` by merging the string.
    config = string_to_config(meta)        
    # create itag, a (ordered) list of frame keys, only for frames with data
    itag = make_channel_tags(meta)
    hasdata = [config.get('DataSet',t,fallback=None) is not None for t in itag]
    itag = [i for (i, hd) in zip(itag, hasdata) if hd]
    
    if np.ndim(index) == 0:
        if read_tags:  #return tags of subscans with data.
            if index == 0:
                return itag
        else:
            index=[index]
    if read_tags: return [itag[i] for i in index]
                    
    for i in index:
        #breakpoint()
        cgtag = itag [i]   
        # all raws of the actual column   
        #pdb.set_trace() 
        #print(cgtag)
        logging.info('tag '+cgtag)
        try:
            datatag = config.get('DataSet',cgtag)
            # then read the header, specially �Points� and �Lines� 
            #ReadDataSetInfo(datatag)    
            npoints = int(config.get(datatag,'Points'))
            nlines = int(config.get(datatag,'Lines' ))
            nbits = int(config.get(datatag,'SaveBits'))
            sign = config.get(datatag,'SaveSign')
            if (int(nbits) != 32) or sign !='Signed':
                raise ValueError
            else:
                fmt = '<l'    
            img = read_datablock(data, npoints, nlines, nbits)
            
            #npoints nlines might be inverted
            Dim0Range = float(config.get(datatag,'Dim0Range'))
            Dim0Min = float(config.get(datatag,'Dim0Min'))
            x = np.arange(npoints) / (npoints-1)  * Dim0Range + Dim0Min
            
            Dim1Range = float(config.get(datatag,'Dim1Range'))
            Dim1Min = float(config.get(datatag,'Dim1Min'))
            y = np.arange(nlines) / (nlines-1) * Dim1Range + Dim1Min
            
            Dim2Range = float(config.get(datatag,'Dim2Range'))
            Dim2Min = float(config.get(datatag,'Dim2Min'))
            img = (img + 2**(nbits-1)) / (2**nbits-1) * Dim2Range + Dim2Min
            # z_value = (z_data + 2^(SaveBits-1)) / (2^SaveBits-1)  * Dim2Range + Dim2Min
            imgdic.append([img,x,y])
            units = [config.get(datatag,'Dim0Unit'),
                        config.get(datatag,'Dim1Unit'),
                        config.get(datatag,'Dim2Unit')]
            
            #if units != ['m','m','m']: 
                #raise NotImplementedError('unknown units in read_nid: ',units)
            #print(units)
            logging.info('units '+units)
            #print(points)
            #ReadBinData()
            logging.info(cgtag+' read')
            #print (cgtag+' read')
        except:
            
            #breakpoint()
            logging.info('option '+cgtag+' not found')
            #print('option '+cgtag+' not found')
            pass
        #breakpoint()
    if len (imgdic)==1: imgdic=imgdic[0] 
        
    return imgdic

def nid_reader(file,index=0,header=False,*args,**kwargs):
    """read .sur binary files. Incompleta."""
    
    from pySurf.readers.nid_reader import read_raw_nid  #,read_nid
    from dataIO.config.make_config import string_to_config
    meta = read_raw_nid(file)[0]
    if header: return meta

    config = string_to_config(meta)
            
    ddic = read_nid(file,*args,**kwargs) # head['Gr0-Ch1']
    #data,x,y = 
    
    return data,x,y

   
def nid_reader2(wfile,header=False):
    """read .nid binary files."""
    head=read_nid(wfile)
    if header: return head

    data,x,y=head['Gr0-Ch1']
    return data,x,y

def read_sur(file,header=False):
    """Convert `res` object returned by `readsur` reader to data,x,y"""
    res = readsur(file,raw = False)
    data,x,y = res.points,res.xAxis,res.yAxis
    del res.points
    del res.xAxis
    del res.yAxis
    return None if header else (data,x.flatten(),y.flatten())

def fits_reader(fitsfile,header=False):
    """ Generic fits reader, returns data,x,y.

    header is ignored. If `header` is set to True is returned as dictionary."""

    a=fits.open(fitsfile)
    head=a[0].header
    if header: return head

    data=a[0].data
    a.close()

    x=np.arange(data.shape[1])
    y=np.arange(data.shape[0])

    return data,x,y
    
    
def test_read_nid(file_name=None):
    from pySurf.data2D_class import Data2D
    from matplotlib import pyplot as plt
    if file_name is None:
        file_name=r'..\test\input_data\AFM\02_test.nid'
    
    datadic = read_nid(file_name)
    data,x,y = datadic['Gr0-Ch1']
    print("read data of shape",data.shape)
    
    d = Data2D(data,x,y,units=['mm','mm','um'],scale=[1000.,1000.,1000000.]) 

    d.plot()
    plt.show()
    return d

def test_reader(file,reader,outfolder=None,infolder=None,**kwargs):
    """called without `raw` flag, return data,x,y. Infolder is taken from file or can be passed (e.g. to point to local test data)."""
    
    import os
    import matplotlib.pyplot as plt
    from  pySurf.data2D import plot_data
    
    if infolder is None:
        infolder=os.path.dirname(file) 
    
    df=os.path.join(infolder,os.path.basename(file))
    res=reader(df,**kwargs)
    header = reader(df,header=True,**kwargs)
    print("reading file %s with reader %s"%(df,reader))
    print("returned values",[r.shape for r in res],header)
    
    plot_data(res[0],res[1],res[2])
    plt.title(os.path.basename(file)+' '+' '.join(["%s=%s"%(k,v) for k,v in kwargs.items()]))
    if outfolder is not None:
        if outfolder == "" : 
            display(plt.gcf()) 
        else: 
            outname=os.path.join(infolder,outfolder,os.path.basename(df))
            os.makedirs(os.path.dirname(outname),exist_ok=True)
            plt.savefig(fn_add_subfix(outname,'','.png'))
    return res,header
    
#used by auto_reader to open according to extension
reader_dic={#'.asc':csvZygo_reader,
            '.csv':csv4D_reader,
            #'.fits':fitsWFS_reader,
            #'.txt':points_reader,
            '.sur':sur_reader,
            '.dat':datzygo_reader,
            '.nid':read_nid}

def auto_reader(wfile):
    """guess a reader for wfile. Return reader routine."""
    ext=os.path.splitext(wfile)[-1]
    try:
        reader=reader_dic[ext]
    except KeyError:
        print ('fileformat ``%s``not recognized for file %s'%(ext,wfile))
        print ('Use generic text reader')
        reader=data_from_txt
        #points_reader  #generic test reader, replace with asciitables

    return reader


if __name__=='__main__':
    """It is based on a tentative generic function read_data accepting among arguments a specific reader.
        The function first calls the data reader, then applies the register_data function to address changes of scale etc.
        This works well, however read_data must filter the keywords for the reader and for the register and
        this is hard coded, that is neither too elegant or maintainable. Note however that with this structure it is
        possible to call the read_data procedure with specific parameters, for example in example below, the reader for
        Zygo cannot be called directly with intensity keyword set to True without making a specific case from the other readers,
        while this can be done using read_data. """

    from .test_readers import testfolder, testfiles
    from pySurf.data2D import plot_data, read_data
    
    tests=[[sur_reader,
    os.path.join(testfolder,r'input_data\profilometer\04_test_directions\05_xysurf_pp_Intensity.sur')
    #,{'center':(10,15)}],[points_reader,
    ,{}],[points_reader,
    os.path.join(testfolder,r'input_data\profilometer\04_test_directions\05_xysurf_pp_Intensity.txt')
    # questo fallisce, perche' delimiter e' " "  e non e' possibile passare l'argomento al reader
    #os.path.join(testfolder,r'input_data\exemplar_data\scratch\110x110_50x250_100Hz_xyscan_Height_transformed_4in_deltaR.dat')
    #,{'center':(10,15)}
    ,{}],
    [csvZygo_reader,
    os.path.join(testfolder,r'input_data\zygo_data\171212_PCO2_Zygo_data.asc')
    ,{'strip':True}
    #,'center':(10,15)}
    ],
    [csvZygo_reader,
    os.path.join(testfolder,r'input_data\zygo_data\171212_PCO2_Zygo_data.asc')
    ,{'intensity':True}]]
    #,{'strip':True,'center':(10,15),'intensity':True}]]

    plt.ion()
    plt.close('all')
    for r,f,o in tests:  #reader,file,options
        plt.figure()
        test_reader(f,r,**o)    


    '''
    for r,f,o in tests:  #reader,file,options
        print ('reading data %s'%os.path.basename(f))
        plt.figure()
        plt.subplot(121)
        plot_data(*r(f))
        plt.title('raw data')
        plt.subplot(122)
        data,x,y=read_data(f,r,**o)
        plot_data(data,x,y)
        plt.title('registered')
        plt.suptitle(os.path.basename(f)+' '+' '.join(["%s=%s"%(k,v) for k,v in o.items()]))
        plt.tight_layout()
        plt.show()
    '''


"""A partire da test_read_sur, trasformato in test_readers.

2020/03/18
"""


'''
import unittest
from unittest import TestCase

testfolder=r'G:\\My Drive\\libraries\\python\\userKov3\\pySurf\\test' #for economy during development, hard coded path.

import os
from format_reader import read_sur
import numpy as np
from pySurf.data2D import plot_data

class TestRead_sur(TestCase):
    def test_read_sur(self):
        """called without `raw` flag, return data,x,y"""
        df=os.path.join(testfolder,r'input_data\\profilometer\\04_test_directions\\02_xprof_m_Intensity.sur')
        res,header=read_sur(df)
        print("returned values",res[0].shape,header)
        #print('test method `xAxis` of returned object')
        self.assertEqual(len(res), 3)
        plot_data(res[0],res[1],res[2])
        return res

if __name__ == '__main__':
    unittest.main()
'''

import os
import sys
import numpy as np
import matplotlib.pyplot as plt
from pySurf.data2D import plot_data
#from format_reader import read_sur
from pySurf.readers.read_sur_files import readsur

#from test_readers import testfolder
from pySurf.data2D import plot_data, read_data
from pySurf.readers.format_reader import points_reader,csvZygo_reader 
from pySurf.readers.format_reader import sur_reader

#ported to/from notebook

#test without unittest, was being inefficient, these run smoother and are more portable.
testfolder = r'..\test'
#testfolder=r'G:\My Drive\libraries\python\userKov3\pySurf\test' #for economy during development, hard coded path.
#testfolder os.path.dirname(__file__) #to set it as relative to this file path.

testfiles = [
r'input_data\profilometer\04_test_directions\05_xysurf_pp_Height.sur',
r'input_data\profilometer\04_test_directions\05_xysurf_pp_Height.txt',
r'input_data\4D\180215_C1S01_RefSub.csv',
r'input_data\4D\180215_C1S01_RefSub.h5',
r'input_data\CCI\01_PCO2S03_00009.sur',
r'input_data\exemplar_data\scratch\110x110_50x250_100Hz_xyscan_Height_transformed_4in_deltaR_matrix.dat',
r'input_data\exemplar_data\scratch\110x110_50x250_100Hz_xyscan_Height_transformed_4in_deltaR_matrix.fits',
r'input_data\exemplar_data\scratch\110x110_50x250_100Hz_xyscan_Height_transformed_4in_deltaR.dat',
r'input_data\fits\reproducibility\181016_01_PCO2S06_1009_08.fits',
r'input_data\fits\reproducibility\181016_01_PCO2S06_1009_08.fits',

r'input_data\newview\105_C1S01.asc',
r'input_data\newview\105_C1S01.dat',
r'input_data\newview\105_C1S01.xyz',

r'input_data\zygo_data\171212_PCO2_Zygo_data.asc',
r'input_data\zygo_data\171212_PCO2_Zygo_data.txt'
]
files=[os.path.join(testfolder,f) for f in testfiles]

  
def test_read_fits(file, raw = False):

    """read a sur file using read_sur_files, that is expected to return a structure
     res.points, .xAxis, .yAxis"""

    res = readsur(file,raw = raw) #in readsur the default is False.
    data,x,y = res.points,res.xAxis,res.yAxis
    del res.points
    del res.xAxis
    del res.yAxis
    header=res #stripped of all data information

    return (data,x,y),header
    
'''	
def csv_points_reader(wfile,*args,**kwargs):
    """Read a processed points file in format x,y,z as csv output of analysis routines."""
    w0=get_points(wfile,*args,**kwargs)
    w=w0.copy()
    x,y=points_find_grid(w,'grid')[1]
    pdata=resample_grid(w,matrix=True)
    return pdata,x,y

def csv_zygo_reader(wfile,*args,**kwargs):
    """Read a processed points file in format x,y,z as csv output of analysis routines."""
    w0=get_points(wfile,*args,**kwargs)
    w=w0.copy()
    x,y=points_find_grid(w,'grid')[1]
    pdata=resample_grid(w,matrix=True)
    return pdata,x,y
'''

def test_readers(tests):
    """test readers as list of 3-element lists, each one corresponding to
    [`reader`,`data`,`parameters_dic`] `reader` is an instrumentReader reader function
    (returns a `Data2D` object)."""
    
    print("Running in path:",os.path.realpath(os.path.curdir) )
    print("File path:", os.path.realpath(sys.argv[0]))

    plt.ion()
    #plt.close('all')
    for r,f,o in tests:  #reader,file,options
        print ('reading data %s'%os.path.basename(f))
        plt.figure()
        plt.subplot(121)
        plot_data(*r(f))
        plt.title('raw data')
        plt.subplot(122)
        data,x,y=read_data(f,r,**o)
        plot_data(data,x,y)
        plt.title('registered')
        plt.suptitle(os.path.basename(f)+' '+' '.join(["%s=%s"%(k,v) for k,v in o.items()]))
        plt.tight_layout()
        plt.show()

            
# %%
if __name__=='__main__':

    # da _instrument_reader: li' va rimossa la registrazione.

    """It is based on a tentative generic function read_data accepting among arguments a specific reader.
        The function first calls the data reader, then applies the register_data function to address changes of scale etc.
        This works well, however read_data must filter the keywords for the reader and for the register and
        this is hard coded, that is neither too elegant or maintainable. Note however that with this structure it is
        possible to call the read_data procedure with specific parameters, for example in example below, the reader for
        Zygo cannot be called directly with intensity keyword set to True without making a specific case from the other readers,
        while this can be done using read_data. """

    pwd = os.path.dirname(os.path.realpath(sys.argv[0]))
    
    tests=[
        [sur_reader,
         os.path.join(pwd,testfolder,r'input_data\profilometer\04_test_directions\05_xysurf_pp_Intensity.sur')
        ,{'center':(10,15)}],
        
        [points_reader,
        os.path.join(pwd,testfolder,r'input_data\profilometer\04_test_directions\05_xysurf_pp_Intensity.txt')
        # questo fallisce, perche' delimiter e' " "  e non e' possibile passare l'argomento al reader
        #os.path.join(testfolder,r'input_data\exemplar_data\scratch\110x110_50x250_100Hz_xyscan_Height_transformed_4in_deltaR.dat')
        ,{'center':(10,15)}],
        
        [csvZygo_reader,
        os.path.join(pwd,testfolder,r'input_data\zygo_data\171212_PCO2_Zygo_data.asc')
        ,{'strip':True,'center':(10,15)}],
        
        [csvZygo_reader,
        os.path.join(pwd,testfolder,r'input_data\zygo_data\171212_PCO2_Zygo_data.asc')
        ,{'strip':True,'center':(10,15),'intensity':True}]
        ]

    test_readers(tests,pwd)







"""Independently from underlying mechanism or preprocessing, provides an interface to access configurations
on files or variables and manipulate them. A configuration is a dictionary
"""

import json
#from pyhocon import ConfigFactory
#https://stackoverflow.com/questions/14050281/how-to-check-if-a-python-module-exists-without-importing-it

import importlib

#if pyhocon is installed, use its method ConfigFactory for `parse`, otherwise use json.load.
if importlib.util.find_spec("pyhocon") is not None:
    from pyhocon import ConfigFactory as parse
    # use configuration files ono HOCON format: The parsed config can be seen as a nested dictionary (with types automatically inferred) where values can be accessed using normal dictionary getter (e.g., conf['a']['b'] or using paths like conf['a.b']) or via the methods get, get_int (throws an exception if it is not an int), get_string, get_list, get_float, get_bool, get_config.
else:
    parse = lambda jfile : json.load(open(jfile,'r'))

    
def dict_from_json(file):
    """just a wrapper around parse, reads a neste dictionary from json file. Return a dictionary."""
    print ('dict_from_json is a wrapper for configuration: please use a more direct reader if you are'+
        ' interested in a specific format (e.g. json.load().')
    return parse(file)

def add_config(confdic,tag,**kwargs):
    """add configuration from variable to confdic.
    Used to migrate old code to new."""
    confdic[tag]={}
    for k,v in kwargs.items():
        confdic[tag][k]=v

def explore_config(config,*strings):
    """ given a list of strings, retrieve the value of the nested config item. """
    for k in strings:
        config = config.get() 
    return config   
        
def read_json_conf(jfile):
    """Read a dictionary from a json file and return a dic. Can include or exclude first level keys.
    
    supercede conf_from_json.
    For `settings` dictionary, each key is a unique name for a configuation,
    The value is a nested dictionary in format {'property_name':<value>}, where property_name 
    is a string and value can be any type (including dictionaries?). 
    
    special keys:
        'default': dictionary of variable names and values that are used for
        every entry in the configuration when these variables are not defined.
        'override': dictionary of variable names and values that override every
            entry in the configuration.
        'includeonly': if set to some list of strings, includes only configurations that are part of the list
        'exclude': list as in includeonly, exclude the listed config, is applied after includeonly.
        """
    
    
    d=parse(jfile)
    #d=json.load(open(jfile,'r'))

    config_settings=d.pop('config_settings',None)    
    default=d.pop('default',None)            
    override=d.pop('override',None)    
    
    if config_settings is not None:
        if 'includeonly' in config_settings:
            if len(config_settings['includeonly'])>0:
                d={k:d[k] for k in config_settings['includeonly'] if k in d}        
        
        if 'exclude' in config_settings:
            #import pdb
            #pdb.set_trace()
            for k in config_settings['exclude']:
                res=d.pop(k,None)
                if res is None:
                    print ("WARNING: json config exclude file not found %s"%k)     
                    
    if default is not None:
        for tt,cc in d.items(): #tag and conf
            for k,dv in default.items():
                cc[k]=cc.get(k,dv)  #or also d[tt][k]=cc.get('k',dv)
       
    if override is not None:
        #print('override')
        for tt,cc in d.items(): #tag and conf
            for k,dv in override.items():
                cc[k]=dv  
                #print('%s override %s %s'%(tt,k,dv))
                
        
                
    return d
    
    
def write_json_conf(jfile,conf):
    """write a configuration to a json file."""
    json.dump(conf, open(jfile,'w'), sort_keys=True, indent=4)  #according to https://stackoverflow.com/questions/7100125/storing-python-dictionaries
    #json.dump(conf,open(jfile,'w'),indent=1)
    
    
    
def conf_from_json(jfile,include=None,exclude=None):
    
    """ Read a dictionary from a json file and return a dic. Can include or exclude first level keys.
    
    For `settings` dictionary, each key is a unique name for a configuation,
    The value is a nested dictionary in format {'property_name':<value>}, where property_name 
    is a string and value can be any type (including dictionaries?). 
    
    If included is set to a list,
    only the keys of first level (configurations) matching names in this list are included.
    In alternative, exclude can be set, leading to include all but these names.
    
    """ 
    
    PRINT ("`conf_from_json` will be eliminated, use instead `read_json_conf(jfile)` ")
        
    d=json.load(open(jfile,'r'))
    #print("%i keys read from file.."%(len(d)))  
    
    if include is not None:
        d={k:d[k] for k in d.keys() if k in include}
    else:
        if exclude is not None:
            for e in exclude:
                try:
                    d.pop(e)
                except (KeyError,ValueError):
                    print("key not found for ",e)
    
    print("----")
    print("included keys: ",list(d.keys()),'\n\nfirst element:\n',d[list(d.keys())[0]])
    print(len(list(d.keys()))," datasets")
    
    return d
    




#2018/11/02 moved to scripts.repeatabilty from WFS_repeatability in POOL\PSD\WFS_repeatability
#v4 2018/09/18 include functions from reproducibility plot. Includes transformations
#  and differences for any number of plots (it was 3).
#v3 added functions all functions from repeatability.
#v1 works, this v2 adds finer control of styles in make_styles with list value:symbol.

import collections

import numpy as np
import matplotlib.pyplot as plt
import os
from cycler import cycler
from itertools import cycle
import pdb
import matplotlib.dates as mdates
import matplotlib.patches as mpatches
from matplotlib.lines import Line2D
import json
import pandas as pd
from plotting.backends import maximize
from pySurf.data2D_class import Data2D
from pySurf.readers.instrumentReader import fitsWFS_reader,matrix4D_reader
from dataIO.fn_add_subfix import fn_add_subfix
"""2019/04/08 from meassy script/repeatability, function that need to be sorted and moved to repeatability. When a function is needed by e.g. a notebook, move it from repeatability_dev to repeatability."""

from utilities.imaging import fitting as fit
from dateutil.parser import parse

from dataIO.outliers import remove_outliers
from plotting.add_clickable_markers import add_clickable_markers2
from pySurf.affine2D import find_affine, find_rototrans
from plotting.multiplots import find_grid_size, subplot_grid
import itertools

from pySurf.data2D_class import align_interactive
from plotting.multiplots import commonscale
from config.interface import conf_from_json

from scripts.repeatability import dcouples_plot

    
def plot_cross_diff_rms(stats):
    """Plots a colormap with values of rms for all cross differences."""
    nfiles=len(stats)
    m_err=np.zeros((nfiles,nfiles))*np.nan
    for j in range(1,nfiles):
        for i in range(j):
            m_err[j,i]=stats['con']['%02i-%02i'%(i+1,j+1)]

    plt.figure()
    plot_data(m_err,range(1,nfiles+1),range(1,nfiles+1),units=['','','$\mu$m'])


    cmap=matplotlib.cm.get_cmap()

    for j in range(1,nfiles):
        for i in range(j):
            """copied from seaborn heatmap"""
            cval = matplotlib.colors.Normalize(vmin=np.nanmin(m_err), 
                                                vmax = np.nanmax(m_err))(m_err[i,j])
            color = cmap(cval)       
            rgb = mpl.colors.colorConverter.to_rgba_array(color)[:, :3]
            rgb = np.where(rgb <= .03928, rgb / 12.92, ((rgb + .055) / 1.055) ** 2.4)
            lum = rgb.dot([.2126, .7152, .0722])
            try:
                l = lum.item()
            except ValueError:
                l = lum
            text_color = ".15" if l > .408 else "w"    
            plt.text(i+1,j+1,'%.3f'%m_err[j,i], color=text_color,
                     horizontalalignment='center',verticalalignment='center')

    plt.xlabel("index of file #1")
    plt.ylabel("index of file #2")
    plt.title("rms of diff for file #1 - #2")


    display(plt.gcf())

# PLOTTING FOR SINGLE SET OF FILES


    




def process_set(flist,m_trans=None,m_arr=None,outfolder=None):
    rmslist=[]       
    
    #plot surfaces, dlist is list of Data2D objects with raw data
    dlist=plot_repeat(flist,
                    None if outfolder is None else os.path.join(outfolder,k+'.png'),dis=dis)
    rms=[d.level(1,1).std() for d in dlist]
    
    if m_trans is None:
        tmp2=dlist
    else:
        tmp2=[d.apply_transform(t) for d,t in zip(dlist,m_trans)]
    
    if m_arr is not None:    
        #plot transformed data for verification
        fig,axes=plt.subplots(xs,ys,sharex='all',sharey='all',squeeze=True)
        axes=axes.flatten()
        maximize()
        for d,ax,m,t in zip(tmp2,axes,m_arr,m_trans):
            plt.sca(ax)
            ll=d.level(4,byline=True)
            ll.plot()
            plt.clim(*remove_outliers(ll.data,nsigma=2,itmax=1,span=True))
            plt.plot(*((t(m).T)),'xr')
        if outfolder is not None:
            plt.savefig(None if outfolder is None else os.path.join(outfolder,k+'_alignment.png'))

    #plot differences with various levelings
    difflist = plot_rep_diff(tmp2,None if outfolder is None else os.path.join(outfolder,k+'.png'),dis=dis)
    rmslist.append([[dd.std() for dd in d] for d in difflist])
    
    if columns is None:
        columns = pd.RangeIndex(0,np.size(rmslist))

    if stats is None: stats = pd.DataFrame(columns=columns)
        
    pstats = pd.DataFrame(columns=stats.columns,data=np.array(rmslist).reshape(1,-1),index=[k])
    
    return pstats
    
def process_set2(flist,names,outname,crop=None,dis=True):
    # from C1S15_4D_DI_calibration_181031.ipynb
    # called with:
    ypix=None #0.127901442999 #mm  
    ytox=220./200.  
    zscale=None #0.6328
    scale=(1000,-1000,0.001)

    outfolder=os.path.dirname(outname)
    outname=os.path.basename(outname)
    
    dlist=[Data2D(*matrix4D_reader(wf2,scale=scale,zscale=zscale,
       ypix=ypix,ytox=ytox,center=(0,0),strip=True),name=n,units=['mm','mm','um']) 
       for wf2,n in zip(flist,names)]
    np.mean(dlist).save(os.path.join(outfolder,outname+'_avg.dat'))
    
    plt.close('all')
    sx,sy=find_grid_size(len(dlist),3)
    f,axes=plt.subplots(sx,sy,sharex=True,sharey=True)
    for ax,d in zip(axes.flatten(),dlist):
        plt.sca(ax)
        d.plot()
        plt.clim(*remove_outliers(d.data,nsigma=2,itmax=3,span=1))
    for ax in axes.flatten()[:len(dlist)-1:-1]:
        f.delaxes(ax)

    maximize()
    plt.pause(0.1)
    plt.tight_layout(rect=[0,0,1,0.9])
    plt.suptitle ('Surfaces')
    if dis:
        display(plt.gcf())
    
    plt.savefig(os.path.join(outfolder,outname+'.png'))
    
    m_repr=[dcouples_plot([dd for dd in dlist])]
    if dis:
        display(plt.gcf())
    plt.savefig(os.path.join(outfolder,'diff_'+outname+'.png'))
    
    if crop is not None:
        dc=[d.crop(*crop).level() for d in dlist]
        m_repr=[dcouples_plot([dd for dd in dc])]
        if dis:
            display(plt.gcf())
        plt.savefig(os.path.join(outfolder,'diff_'+outname+'_crop.png'))
       
    return dlist
    
## FUNCTIONS TO HANDLE MULTIPLE FILES    
    
def make_styles(s1,cf,argsdic,legshow=None):
    """given a dataframe s1, associate to each element in s1 a plotting style in form of dictionary
        according to values of columns for s1 indices in cf.
        argsdic has the form {'graphic_property':{'colkey':[style1,style2]}},
        where col_to is the column index that is used to determine the graphic style.
        it is done in this format {gp:{ck:stylst}} rather than the more intuitive {ck:{gp:stylst}}
        to conform to the standard {gp:values}.
        Note that for each graphic_property, there should be a single key for the associated dictionary 
        {kc:stylst}
        Also, the entire dataframe s1 is passed instead of its index, so that a function on s1
        columns can be added subsequently (maybe this is useless and all processing should be
        handled on cf, that can also be =s1 if this contains already all information).
        Legshow is a list with indices (in form of `colkey`) that tells
        which legend to plots (locations are determined in increasing values of loc starting from 1, with legend plotted in order as 
        in argsdic). Set to empty string to disable plotting, legends can
        be plotted at later time using legenddic.
        """
    # TODO:
    # - come posso dare un doppio stile (e.g. red circles vs blue triangles
    # - aggiustare legende per plottare solo simboli o solo linee con stili comuni inclusi 
    # - how to apply a legend different than value (e.g. chiller on/off instead of 0/1)
    
    stylelist=[]
    #stylelist=[{} for i in range(len(s1.index))]  #[{}]*len doesn't work, it creates n copies of same dictionary, so any inplace change to an element is reflected to other elements
    if legshow is None:
        legshow=[list(argsdic[k].keys())[0] for k in list(argsdic.keys()) if isinstance(argsdic[k], collections.Mapping)]
    legenddic={}
    pos=1
    j=0  #ugly way to check when it's first iteration and creating item    
    for k,v in argsdic.items():   #iterate over column tags associated with the property
        if isinstance(v, collections.Mapping): #dictionary
            assert len(v.keys())==1            
            import pdb
            for p,c in v.items(): #iterate over properties, anyway this will always be a single key,
                newkeys=np.unique(cf[p].values)
                if isinstance(c, collections.Mapping): #dictionary
                    sd={kk:{k:vv} for kk,vv in c.items()}
                elif isinstance(c, list):            
                    #so there might be neater ways to unpack
                    p_cycle=cycler(k,c)

                    #builds a dictionary with unique values as keys and property value as value
                    sd={}
                    for nk,sty in zip(newkeys,cycle(p_cycle)):
                        sd[nk]=sty
                    #{'LSKG': {'color': 'r'}, 'VC': {'color': 'g'}}
                    #{'CylRef': {'marker': 'x'}, 'PCO1.2S01': {'marker': 'o'}, 'PCO1S23': {'marker': '+'}}
                    #print(sd)
                    
            # builds the return value stylelist (list of graphic properties
            # associate a style to each row of stats
            for i,cc in enumerate(cf[p].values):
                #stylelist,legenddic=test_make_styles(sc,cc,   #doesn't plot others
            #{"color":{"operator":{'KG':'r','LS':'b'}}})
                #stylelist[i].update(sd.get(None,sd.get(cc,{'marker':'','linestyle':''})))
                #pdb.set_trace()
                #stylelist[i]= 
                if isinstance(v, collections.Mapping): #dictionary
                    s = sd.get(cc,sd.get(None,None)) #set to key if in sd, otherwise to the graph prop dictionary for None if it was
                    #set, or set the style to None (exclude in plot_historical) if not.   
                else:
                    s = {k:v}
                
                if j == 0:
                    stylelist.append(s)
                else:
                    stylelist[i] = None if (stylelist[i] is None or s is None) else {**stylelist[i],**s}        
            j=1   

            #pdb.set_trace()
            #make a dictionary of legends for each of the keys in stylelist
            legenddic[p]=[[sd[t],t] if t is not None else [sd[t],'Other'] for t in sd.keys()]
            #handles, labels = plt.gca().get_legend_handles_labels() # get existing handles and labels
            #empty_patch = mpatches.Patch(color='none', label='Extra label') # create a patch with no color
            #handles.append(empty_patch)  # add new patches and labels to list
            #labels.append("Extra label")

            #plt.legend(handles, labels) # apply new handles and labels to plot
            handles=[Line2D([], [], label= vv[1], **(vv[0])) for vv in legenddic[p]]
            labels=[vv[1] for vv in legenddic[p]]
            #handles=[v[0] for v in ]
            
            if p in legshow:
                plt.gca().add_artist(plt.legend(handles,labels,title=p,loc=pos))
            pos=pos+1
            
        else:
            for i in range(len(cf.index)):
                s = {k:v}
                if j == 0:
                    stylelist.append(s)
                else:
                    stylelist[i] = None if (stylelist[i] is None or s is None) else {**stylelist[i],**s}        
            j=1   
        
    return stylelist,legenddic
        
def test_make_styles(s1,cf,kwargs):
    plt.clf()
    stylelist,legenddic=make_styles(s1,cf,kwargs)
    print("styledic(styles for each line):\n%s \nlegenddic(dictionary of legends):\n%s\n"%
          (stylelist,legenddic))
    display(plt.gcf())
    print ("\n")
    return stylelist,legenddic


def build_database(confdic,outfolder=None,columns=None,dis=False,
                   find_transform=find_rototrans):
    """Process all files and return stats database in form of pd.DataFrame from confdic. 
    plot_repeat and plot_rep_diff are used to read fits files and calculate rmss.
    If outfolder is provided, output plots are generated in functions.
    Expect an index for columns (can be multiindex.)
    If dis is set, display (set to True for single dataset, False for multiple config processing."""
    
    def make_dates(stats):
        """returns a dataframe dates build with indices from stats and value """
        dates=[]
        for k in stats.index:
            try:
                date=confdic[k]['date']
            except KeyError:
                date=k.split('_')[0]
            dates.append(parse(date,yearfirst=True).date()) 

        return pd.Series(data=dates,index=stats.index)
    
    
   #check existence of files before starting processing
    for k,conf in confdic.items():
        flist=[os.path.join(conf['infolder'],f)for f in conf['files']]
        if not np.all([os.path.exists(f) for f in flist]):
            print (k," has some files that cannot be found:")
            print (conf['infolder'])
            print (conf['files'])
            for f in flist:
                open(f).close()  #raise exception if file doesn't exist
    
    stats = None
    #real processing
    for i,(k,conf) in enumerate(confdic.items()):
        print("Process data %i/%i"%(i+1,len(confdic.items())))
        
        #######
        rmslist=[]
        
        #outfile=os.path.join(outfolder,k)
        flist=[os.path.join(conf['infolder'],f)for f in conf['files']]
        
        #plot surfaces, dlist is list of Data2D objects with raw data
        dlist=plot_repeat(flist,
                        None if outfolder is None else os.path.join(outfolder,k+'.png'),dis=dis)
        rms=[d.level(1,1).std() for d in dlist]
        
        #align if necessary:
        transmode=conf.get('trans',None)
        if transmode=='none' or transmode == "":  #do nothing
            tmp2=dlist
        else:
            plt.close('all')
            if transmode == 'interactive':
                print ("""interactively set markers on each axis with CTRL+LeftMouse. 
                Remove markers with CTRL+right_mounse, ENTER when ready.""")
                #run interactive alignment 
                m_arr,m_trans=align_interactive(dlist,find_transform=find_transform)

                #incorporate markers in config and save backup copy 
                conf['trans']='config'   # alignment mode, read from config
                conf['mref']=m_arr[0].tolist()    #reference markers to aligh to
                conf['markers']=[m.tolist() for m in m_arr]    #markers saved as lists
                json.dump({k:conf},open(os.path.join(outfolder,k+'_out.json'),'w'),indent=0)            

            else:
                if transmode == 'config':    #gets markers from configfile
                    pass  #conf is already set
                elif transmode == 'file':
                    d=json.load(open(conf['markers'],'r'))
                    print(list(d.keys()),'\n\n',d[list(d.keys())[0]])
                    print(len(list(d.keys()))," datasets")  
                    k,conf=list(d.items())[0]                    
                else:
                    raise ValueError
                m_arr=[np.array(m) for m in conf['markers']]
                mref=conf['mref']
                m_trans=[find_transform(m,mref) for m in m_arr]  
                        
            #make and plots transformed (aligned) data to verify feature alignment
            tmp2=[d.apply_transform(t) for d,t in zip(dlist,m_trans)]

            #plot transformed data for verification
            fig,axes=plt.subplots(xs,ys,sharex='all',sharey='all',squeeze=True)
            axes=axes.flatten()
            maximize()
            for d,ax,m,t in zip(tmp2,axes,m_arr,m_trans):
                plt.sca(ax)
                ll=d.level(4,byline=True)
                ll.plot()
                plt.clim(*remove_outliers(ll.data,nsigma=2,itmax=1,span=True))
                plt.plot(*((t(m).T)),'xr')
            plt.savefig(None if outfolder is None else os.path.join(outfolder,k+'_alignment.png'))
        
        #yield dlist
        #plot differences with various levelings
        difflist = plot_rep_diff(tmp2,None if outfolder is None else os.path.join(outfolder,k+'.png'),dis=dis)
        rmslist.append([[dd.std() for dd in d] for d in difflist])
        
        if columns is None:
            columns = pd.RangeIndex(0,np.size(rmslist))

        if stats is None: stats = pd.DataFrame(columns=columns)
            
        pstats = pd.DataFrame(columns=stats.columns,data=np.array(rmslist).reshape(1,-1),index=[k])
            
        #### fuori routine
        
        stats=stats.append(pstats)
        dates=make_dates(stats)

    return stats, dates
    
def plot_historical(dates,stats,col=None,yrange=[0,0.25],styarr=None,*args,**kwargs):
    """plot historical values as scatter plot indicating outliers with arrow and value.  
    dates and stats are data frames with same indices. 
    Plot all data matching column col.
    styarr is a list of graphic styles with same length as dates to apply to each element
    plot. Can be created on the base of data content with make_styles."""
    
    plt.ylim(*yrange)
    al=(plt.ylim()[1]-plt.ylim()[0])*0.05  #arrow length
    tyoff=3*al  #vertical offset for text in outliers plots 
    
    m_cycle = cycler(marker=['s', 'x', 'v', '+', '^', 'o' ])()
    #print(kwargs)
    #
    if styarr is None:
        styarr=[m for i,m in zip(range(len(dates)),cycle(m_cycle))]
        #styarr=[{} for i in range(len(dates))]
    #pdb.set_trace()
    styarr=[a if a is None else (a if 'marker' in a.keys() else {**a,**m}) for a,m in zip(styarr,cycle(m_cycle))]
    
    #sarebbe meglio fare passare gia' ordinati
    isort=np.argsort(dates.values)
    dates=dates[isort]
    stats=stats.iloc[isort]
    styarr=[styarr[i] for i in isort]
    
    oldate=None
    #plt.figure(figsize=([11.,6.35]))
    for i,sty in zip(range(len(dates)),styarr):
        if sty is not None:
            n=stats.index[i]
            #print(n)
            #pdb.set_trace()
            points=(stats[col] if col is not None else stats).loc[n]
            out=[p for p in points if p >= plt.gca().get_ylim()[1]]
            inr=[p for p in points if p < plt.gca().get_ylim()[1]]
            if len(out) > 0:
                #if it is out of y range, plot arrow and text 
                if dates[i] == oldate: #if more than one data same date
                    tyoff=tyoff+al     #set offset for text
                oldate=dates[i]
                plt.plot(dates[i],plt.gca().get_ylim()[1]-al, 'k', marker=r'$\uparrow$', markersize=10)
                
                t=plt.text(mdates.date2num(dates[i]),plt.gca().get_ylim()[1]-tyoff,'%3.2g'%np.nanmax(out))
                plt.plot(dates[i],plt.gca().get_ylim()[1]- 2*al, 'k',**sty)
                #print(a['con'][n])
            if len(inr)>0:
                plt.plot(np.repeat(dates[i],len(inr)),inr,label=n,**sty)

    plt.xticks(rotation=60)
    plt.legend(bbox_to_anchor=(1.04,1), loc="upper left")
    
    plt.ylabel('rms (um)')
    plt.gcf().subplots_adjust(bottom=0.15,right=0.7)
    plt.grid(axis='y')
    #display(plt.gcf())

##rep_stats and conf_stats from tackbonding notebooks.

def rep_stats(flist,columns=None,dis=True):
    """Analyzes a list of files with possible transformations.
        returns repeatability statistics in form of pdDataFrame. """
    rmslist=[]
    
    #plot surfaces, dlist is list of Data2D objects with raw data
    dlist=plot_repeat(flist,
                    None if outfolder is None else os.path.join(outfolder,k+'.png'),dis=dis)
    rms=[d.level(1,1).std() for d in dlist]

    tmp2=[d if t is None else t(d) for d,t in zip(dlist,m_trans)]
    
    #align if necessary:    
    if transmode=='none' or transmode == "":  #do nothing
        tmp2=dlist
    else:
        plt.close('all')
        if transmode == 'interactive':
            print ("""interactively set markers on each axis with CTRL+LeftMouse. 
            Remove markers with CTRL+right_mounse, ENTER when ready.""")
            #run interactive alignment 
            m_arr,m_trans=align_interactive(dlist,find_transform=find_affine)          

        else:
            if transmode == 'file':
                #read markers from file that some 
                raise NotImplementedError
            else:
                raise ValueError
            m_arr=[np.array(m) for m in conf['markers']]
            mref=conf['mref']
            m_trans=[find_transform(m,mref) for m in m_arr]  

        #make and plots transformed (aligned) data to verify feature alignment
        tmp2=[d.apply_transform(t) for d,t in zip(dlist,m_trans)]

        #plot transformed data for verification
        fig,axes=plt.subplots(xs,ys,sharex='all',sharey='all',squeeze=True)
        axes=axes.flatten()
        maximize()
        for d,ax,m,t in zip(tmp2,axes,m_arr,m_trans):
            plt.sca(ax)
            ll=d.level(4,byline=True)
            ll.plot()
            plt.clim(*remove_outliers(ll.data,nsigma=2,itmax=1,span=True))
            plt.plot(*((t(m).T)),'xr')
        plt.savefig(None if outfolder is None else os.path.join(outfolder,k+'_alignment.png'))

    #yield dlist
    #plot differences with various levelings
    difflist = plot_rep_diff(tmp2,None if outfolder is None else os.path.join(outfolder,k+'.png'),dis=dis)
    rmslist.append([[dd.std() for dd in d] for d in difflist])

    if columns is None:
        tags=['%02i-%02i'%(i,j) for i,j in itertools.combinations(range(1,nfiles+1),2)]
        columns=pd.MultiIndex.from_product([['raw','cyl','con','leg'],tags],names=['terms removed','files#'])
        #columns = pd.RangeIndex(0,np.size(rmslist))
    
    #pdb.set_trace()
    cstats = pd.DataFrame(columns=columns,data=np.array(rmslist).reshape(1,-1),index=[k])

    return cstats




def conf_stats(k,conf,columns=None,dis=True):
    """Analyzes a single configuration (as list of files + processing settings).
        returns repeatability statistics in form of pdDataFrame. """
    rmslist=[]

    #outfile=os.path.join(outfolder,k)
    flist=[os.path.join(conf['infolder'],f)for f in conf['files']]
    transmode=conf.get('trans',None)
    
    
    
    #plot surfaces, dlist is list of Data2D objects with raw data
    dlist=plot_repeat(flist,
                    None if outfolder is None else os.path.join(outfolder,k+'.png'),dis=dis)
    rms=[d.level(1,1).std() for d in dlist]

    tmp2=[d if t is None else t(d) for d,t in zip(dlist,m_trans)]
    
    #align if necessary:    
    if transmode=='none' or transmode == "":  #do nothing
        tmp2=dlist
    else:
        plt.close('all')
        if transmode == 'interactive':
            print ("""interactively set markers on each axis with CTRL+LeftMouse. 
            Remove markers with CTRL+right_mounse, ENTER when ready.""")
            #run interactive alignment 
            m_arr,m_trans=align_interactive(dlist,find_transform=find_affine)

            #incorporate markers in config and save backup copy for the configuration. 
            # This is just temporary ass saver in case of crash, final conf is anyway
            # saved outside for all configurations.
            conf['trans']='config'   # alignment mode, read from config
            conf['mref']=m_arr[0].tolist()    #reference markers to aligh to
            conf['markers']=[m.tolist() for m in m_arr]    #markers saved as lists
            json.dump({k:conf},open(os.path.join(outfolder,k+'_out.json'),'w'),indent=0)            

        else:
            if transmode == 'config':    #gets markers from configfile
                pass  #conf is already set
            elif transmode == 'file':
                d=json.load(open(conf['markers'],'r'))
                print(list(d.keys()),'\n\n',d[list(d.keys())[0]])
                print(len(list(d.keys()))," datasets")  
                k,conf=list(d.items())[0]                    
            else:
                raise ValueError
            m_arr=[np.array(m) for m in conf['markers']]
            mref=conf['mref']
            m_trans=[find_transform(m,mref) for m in m_arr]  

        #make and plots transformed (aligned) data to verify feature alignment
        tmp2=[d.apply_transform(t) for d,t in zip(dlist,m_trans)]

        #plot transformed data for verification
        fig,axes=plt.subplots(xs,ys,sharex='all',sharey='all',squeeze=True)
        axes=axes.flatten()
        maximize()
        for d,ax,m,t in zip(tmp2,axes,m_arr,m_trans):
            plt.sca(ax)
            ll=d.level(4,byline=True)
            ll.plot()
            plt.clim(*remove_outliers(ll.data,nsigma=2,itmax=1,span=True))
            plt.plot(*((t(m).T)),'xr')
        plt.savefig(None if outfolder is None else os.path.join(outfolder,k+'_alignment.png'))

    #yield dlist
    #plot differences with various levelings
    difflist = plot_rep_diff(tmp2,None if outfolder is None else os.path.join(outfolder,k+'.png'),dis=dis)
    rmslist.append([[dd.std() for dd in d] for d in difflist])

    if columns is None:
        tags=['%02i-%02i'%(i,j) for i,j in itertools.combinations(range(1,nfiles+1),2)]
        columns=pd.MultiIndex.from_product([['raw','cyl','con','leg'],tags],names=['terms removed','files#'])
        #columns = pd.RangeIndex(0,np.size(rmslist))
    
    #pdb.set_trace()
    cstats = pd.DataFrame(columns=columns,data=np.array(rmslist).reshape(1,-1),index=[k])

    return cstats
    
    
#CONFIGURATION HANDLING



def stats_from_json(jfile,include=None,exclude=None,outfolder=None,columns=None):
    
    """ emulate obsolete function using new functions 
     json->conf (conf_from_json) + conf->stats (build_database).""" 

    confdic = conf_from_json(jfile,include,exclude) 
    stats,dates=build_database(confdic,outfolder=outfolder,columns=columns)
    plt.title('rms after conical correction for different samples')
    display(plt.gcf())
    return stats,dates
    
"""
EXAMPLE OF CONFIG.JSON

{
"180719_CylRef_repeat": {
"infolder": "tests\\input_data\\180719_Repeatibility",
"files": [
"180719_01_CylRef_01.fits",
"180719_02_CylRef_01.fits",
"180719_03_CylRef_01.fits"
],
"outname": "testset01",
"sample": "CylRef",
"testtype": "repeat",
"trans": "",
"operator": "VC",
"date": "2018/07/19",
"markers": []
},

"180719_CylRef_reprod": {
"infolder": "tests\\input_data\\180719_Repeatibility",
"files": [
"180719_04_CylRef_01.fits",
"180719_05_CylRef_01.fits",
"180719_06_CylRef_01.fits"
],
"outname": "testset02",
"sample": "CylRef",
"testtype": "reprod",
"trans": "",
"operator": "LS",
"date": "2018/07/19",
"markers": []
}
}
"""   
   
if __name__=="__main__":
    
    plt.ion()
    #01
    outfolder=os.path.join(r'tests','WFS_repeatability') #output 
    jfile=os.path.join(outfolder,r'test_settings.json')
    include=['180206_c1s03_gentexcut_', '180206_c1s04_gentexcut_'] #limits analysis to only two files
    include=None

    #CALCULATE ALL!
    confdic=conf_from_json(jfile,include=include)
    mi=pd.MultiIndex.from_product([['raw','cyl','con','leg'],['2-1','3-1','3-2']],names=['terms removed','files#'])
    stats, dates = build_database(confdic,outfolder,columns=mi)
    
    plt.figure(figsize=((8,4)))
    plt.title('rms after conical correction for different samples')
    display(plt.gcf())

    plot_historical(dates,stats,'con',outfolder=outfolder)
    display(stats*1000)
    #------------------
    #outfolder=r'results\test_reptester\WFS_repeatability'
    
    #plot_historical(dates,stats,'con',yrange=[0,0.25],color={"operator":['r','g','b']})
    #plot_historical(dates,stats,col,yrange=[0,0.25],color={"operator":cycler})
    #plot_historical(dates,stats,col,yrange=[0,0.25],color={"operator":{'KG':'r','VC':'g','LS':'b',None:'k']}) #is this the most general? We said I want to control only one graphical prop per category,
    #however I might want to set common extra properties. Can this be done with cyclers?
    






import sys
import numpy as np
import sys
sys.ps1 = 'SOMETHING'
from matplotlib import pyplot as plt


#2017/08/03 made some adjustments, mostly in interactivity. However plt.ion/ioff don't update immediately,
#  it is necessary to close the window.

def add_clickable_markers(fig=None,ax=None,key='enter',propertyname='markers',modkey='control',hold=False):
    """Makes pickable the image (image only) in the current figure.
    add a list `propertyname` to the figure. if axis is passed, markers are added only to specific axis. This way it is possible to associate differently named markers to 
        different axis in same figure.
        
        Deprecated, use add_clickable_markers2"""
    
    #useful references  
    #http://astronomy.nmsu.edu/holtz/a575/ay575notes/node6.html
    #https://matplotlib.org/examples/event_handling/keypress_demo.html
    def press(event):
        #print('press', event.key)
        sys.stdout.flush()
        if event.key == key:
            fig.canvas.mpl_disconnect(cid1)
            fig.canvas.mpl_disconnect(cid2)
            #fig.canvas.stop_event_loop()
            print (plt.isinteractive())
            plt.interactive(interact_store)  #this sets interactive to original status, but the behaviour doesn't update (if set to ON you still need to close the figure)
            #plt.show(block=False)
            print (plt.isinteractive())
            fig.exit=True  #experimental for loop
        
    def onclick(event):
        button=event.button
        x=event.xdata
        y=event.ydata
        fig=event.canvas.figure
        cax=event.inaxes
        
        if ax is None or cax==ax:         
            if button==1: 
                if event.key is not None:
                    if event.key == modkey:
                        #print (id(fig),propertyname)
                        l=getattr(fig,propertyname)
                        l.append([event.xdata, event.ydata])
                        #add point
                        cax.set_autoscale_on(False)
                        cax.plot(x,y,'or',picker=5,markerfacecolor=None)
            if button==3:
                if event.key is not None:
                    if event.key == modkey:
                        l=getattr(fig,propertyname)
                        if len(l)>0:
                            dist=((np.array(l)-(x,y))**2).sum(axis=1)
                            pop=l.pop(np.argmin(dist))
                            dist2=[(np.array((pop[0]-a.get_xdata(),pop[1]-a.get_ydata()))**2).sum()  for a in ax.lines]
                            cax.lines.pop(np.argmin(dist2))
        
        plt.draw()
        #print 'button=%d, x=%d, y=%d, xdata=%f, ydata=%f'%(
        #event.button, event.x, event.y, event.xdata, event.ydata)
    
    def handle_close(evt):
        print('Closed Figure!')
        fig.canvas.mpl_disconnect(cid3)

    #to disconnect:
    #fig.canvas.mpl_disconnect(cid)
    """
    def on_pick(event):
        # not used. this is an alternative way, it automatically manage the selection of the plotted point.
        import matplotlib
        thisline = event.artist
        if isinstance(thisline,matplotlib.lines.Line2D):
            #code to remove point from markers list
            thisline.remove()
            plt.show()
    """
    
    if fig is None:
        fig=plt.gcf()
    interact_store=plt.isinteractive()
    #plt.show(block=True) #This makes impossible to select the points because it blocks events handling.
    #plt.show(block=False) #This works to select the points, but it doesn't release afterwards.

    if hold:
        print ('hold')
        plt.ioff()
    else:
        plt.ion()  #this works in setting interactive mode off independently on starting status. However it sets the status forever, setting ioff here and ion at the end alters
    # the status variable, but it is not applied to the window.
    # so it must be set off here if you want to proceed at the end, but it is necessary to 
    # close the windows to continue.
    
    #fig.canvas.start_event_loop(20)  #this is useful only for timed events (ginput?) documentation is
        #not  very clear to this regar, but it's clearly useless here.
    #add a list property (empty) with the chosen name for markers,
    #if property already exist it will automatically append.    
    if not hasattr(fig,propertyname):
        setattr(fig,propertyname,[])
        #setattr(fig.getattr(propertyname),'ax')
    fig.exit=False #experimental for loop
    #ax.imshow(im)
    #ax.autoscale(False)
    cid1 = fig.canvas.mpl_connect('button_press_event', onclick)
    cid2 = fig.canvas.mpl_connect('key_press_event', press)
    cid3 = fig.canvas.mpl_connect('close_event', handle_close)
        
    return fig

def add_clickable_markers2(ax=None,key='enter',propertyname='markers',modkey='control',hold=False):
    """Enable adding markers by point and click to the passed or current axis. Returns axis.
    
    Markers are added by mouse left click + modkey and are stored in a list added
        to the axis as a property `propertyname`. They can be removed by modkey
        + mouse right click.
        `key` disconnect user interface stopping point and click operations.
    Markers with different propertyname can be associated to same axis. 
    hold is implemented in experimental way, only one axis should hold. Not sure
        what happens otherwise. Even with this working version there is flickering.

    This is updated version of add_clickable_markers, associate with axis, 
        not with figure so it is safe to use on subplots.
    ex:
    ax1=add_clickable_markers2()    #use current axis
    ax1=add_clickable_markers2(ax=ax1)   #use a specific axis
    print (ax1.markers)  #run this after markers are added
    
    2018 vcotroneo@cfa.harvard.edu
    """

    # 2017/10/26 it is more appropriate to associate markers with axis,
    # not with figures, so it will be possible to handle subplots.

    # useful references
    # http://astronomy.nmsu.edu/holtz/a575/ay575notes/node6.html
    # https://matplotlib.org/examples/event_handling/keypress_demo.html
    
    ## routines to be called on events
    ## Note that events and associations belong to figure, the axis is obtained with
    ##     event.inaxes
    def press(event):   #called on keyboard press event
        #print('press', event.key)
        cax=event.inaxes
        #print("cax:","cax is None!" if cax is None else cax)
        #print("*")
        if cax is None:
            cax=plt.gca() #if cursor was not on any axis when key press
        sys.stdout.flush()
        if event.key == key:
            fig=cax.figure
            print ("disconnecting cids:")
            #print(cids[:2]) #2017/10/26 not sure where cid3 is taken, maybe it works, added other cids.
            print("----")
            fig.canvas.mpl_disconnect(cids[0])   #w
            fig.canvas.mpl_disconnect(cids[1])
            #fig.canvas.stop_event_loop()
            print ("isinteractive?",plt.isinteractive())
            plt.interactive(interact_store)  #this sets interactive to original status, but the behaviour doesn't update (if set to ON you still need to close the figure)
            #plt.show(block=False)
            print ("isinteractive?",plt.isinteractive())  
            if hasattr(fig,'exit'):
                fig.exit=True
        
    def onclick(event):     #called on mouse button press event
        button=event.button
        x=event.xdata
        y=event.ydata
        #fig=event.canvas.figure
        cax=event.inaxes
        
        if ax is None or cax==ax:         
            if button==1: 
                if event.key is not None:
                    if event.key == modkey:
                        #print (id(fig),propertyname)
                        l=getattr(cax,propertyname)  ##getattr(fig,propertyname)
                        l.append([event.xdata, event.ydata])
                        #add point
                        ax.set_autoscale_on(False)
                        ax.plot(x,y,'or',picker=5,markerfacecolor=None)
            if button==3:
                if event.key is not None:
                    if event.key == modkey:
                        l=getattr(cax,propertyname)  ##getattr(fig,propertyname)
                        if len(l)>0:
                            dist=((np.array(l)-(x,y))**2).sum(axis=1)
                            pop=l.pop(np.argmin(dist))
                            dist2=[(np.array((pop[0]-a.get_xdata(),pop[1]-a.get_ydata()))**2).sum()  for a in ax.lines]
                            ax.lines.pop(np.argmin(dist2))
        
        plt.draw()
        #print 'button=%d, x=%d, y=%d, xdata=%f, ydata=%f'%(
        #event.button, event.x, event.y, event.xdata, event.ydata)
    
    def handle_close(evt):
        print('Closed Figure!')
        print('cids:')
        print(cids) #2017/10/26 not sure where cid3 is taken, maybe it works, added other cids.
        print("#---")
        #cids are created when function is run. are then they stored in handle_close?
        # to test: call function, then disconnect cids1 and 2 by enter key,
        # then close to call this routine that prints cids.
        # if they are defined, they are retained by function call. 
        for c in cids:
            evt.canvas.mpl_disconnect(c)
    
    # add property `propertyname` as empty list and associate event routines
    # to UI actions that modify list of markers.
    if ax is None:
        ax=plt.gca()
    else:
        if isinstance (ax,plt.Figure):
            raise ValueError   #old implementation was accepting figure
    
    interact_store=plt.isinteractive()
    """
    if hold:  #for test and debuffing
        print ('hold')
        plt.ioff()
    else:
        plt.ion()  
    """
    
    if not hasattr(ax,propertyname):
        setattr(ax,propertyname,[])
        
    fig=ax.figure
    # Deactivate the default keymap
    #fig.canvas.mpl_disconnect(fig.canvas.manager.key_press_handler_id)
    
    cids=[
    fig.canvas.mpl_connect('button_press_event', onclick),
    fig.canvas.mpl_connect('key_press_event', press),
    fig.canvas.mpl_connect('close_event', handle_close)
    ]
    
    print("cids established:\n",cids)
    
    def get_key():
        """proper way to implement press key with yield.
        see 
        https://gist.github.com/tacaswell/4545013"""
        pass
        
    if hold:
        setattr(fig,'exit',False)
        #i=0 #safety
        while not(fig.exit): # and i < 500:
            #i=i+1
            plt.pause(0.01)
        print ("exit")  #,i)
    return ax

class modify_markers(object):
    def __init__(self):
        self.fig = plt.figure()
        self.ax = self.fig.gca()
        self.bound_keys = []
        self.bound_cid = {}

    def add_step_through(self, gen, key):
        
        key = key[0] # make a single char
        if key in self.bound_keys:
            raise RuntimeError("key %s already bound"%key)
        first_data = gen.next()
        self.ax.plot(first_data)
        self.fig.canvas.draw()
        self.bound_keys.append(key)
        def ontype(event):
            if event.key == key:
                try:
                    self.ax.plot(gen.next())
                    self.fig.canvas.draw()
                except StopIteration:
                    self.fig.canvas.mpl_disconnect(self.bound_cid[key])
                    del self.bound_cid[key]
                    self.bound_keys.remove(key)

        #self.bound_cid[key] = self.fig.canvas.mpl_connect('key_press_event', ontype)) 

 
class push_to_advance(object):
    def __init__(self):
        self.fig = plt.figure()
        self.ax = self.fig.gca()
        self.bound_keys = []
        self.bound_cid = {}

    def add_step_through(self, gen, key):
        key = key[0] # make a single char
        if key in self.bound_keys:
            raise RuntimeError("key %s already bound"%key)
        first_data = gen.next()
        self.ax.plot(first_data)
        self.fig.canvas.draw()
        self.bound_keys.append(key)
        def ontype(event):
            if event.key == key:
                try:
                    self.ax.plot(gen.next())
                    self.fig.canvas.draw()
                except StopIteration:
                    self.fig.canvas.mpl_disconnect(self.bound_cid[key])
                    del self.bound_cid[key]
                    self.bound_keys.remove(key)

        self.bound_cid[key] = self.fig.canvas.mpl_connect('key_press_event', ontype)

def test_interac():
    """2018/09/05 try to solve with new idea of calling add_clickable_markers2 as a function that returns a value only
    when a key is pressed. to do this the interactive mode of matplotlib must be off.
    This works perfectly if imported from ipython shell, but if 
    called in qtconsole from jupyter notebook, doesn't disable
    interactive mode."""
    
    plt.ioff()
    plt.figure()
    plt.plot([1, 2, 3], [3, -3, 6])
    plt.show()
    plt.ion()

    
def test_interac2():
    """This probably works better than anything else. Still has flickering, but works from both ipython and jupyter. 
    
    For some strange reason, pause and coord must be
    made global, while there is no need for cid.
    Maybe because global and pause are in functions linked to mpl,
    while remove_cids is independent?"""

    
    def remove_cids():
        for c in cids:
            fig.canvas.mpl_disconnect(c)        
    
    def onclick(event):
        global coords,pause
        coords.append((event.xdata, event.ydata))
        print(pause,len(coords),coords)
        if (len(coords)==2):
            pause = False
            remove_cids()
    
    def handle_close(evt):
        remove_cids()

    global coords,pause
    coords = []
    pause = True

    fig, ax = plt.subplots()
    plt.plot([1, 2, 3], [3, -3, 6])
    plt.show() 
    
    cids=[
        fig.canvas.mpl_connect('button_press_event', onclick),
        #fig.canvas.mpl_connect('key_press_event', press),
        fig.canvas.mpl_connect('close_event', handle_close)
        ]
    print (cids)
 
    while pause:
        plt.pause(0.01)
    
    return coords

def test_multimarkers():
    """plots two datasets in subplots, return axis with attached clickable markers
    (CTRL+button). return axis ax1 and ax2 with markers properties attached named respectively 
    ax1.m1 and ax2.m2 """

    a = np.random.random(20).reshape((5, 4))
    b = -a
    ax1 = plt.subplot(121)
    plt.imshow(a, interpolation='none')
    ax2 = plt.subplot(122)
    plt.imshow(b, interpolation='none')
    add_clickable_markers2(ax=ax1, propertyname='m1')
    f = add_clickable_markers2(ax=ax2, propertyname='m2')
    plt.show()
    return ax1, ax2


def test_loop():
    """test for first version add_clickable_markers (on figure, deprecated)."""
    # doesn't work in spyder 2018/06/04

    plt.clf()
    a = np.random.random(25).reshape(5, 5)
    plt.imshow(a)
    print('add points with CTRL+left button, remove closest with CTRL+right. retax() prints the points plotted on current axes, f.markers print marker list.')

    f = add_clickable_markers() # has experimental property .exit()
    plt.show()
    # ax=plt.gca()
    print(f.markers)

    import time
    while not (f.exit):
        time.sleep(0.1)


def test_add_clickable_markers(a):

    plt.clf()
    plt.imshow(a)

    f = add_clickable_markers()
    plt.show()
    # ax=plt.gca()
    print(f.markers)
    return f


def test_multiaxis(ax1, ax2, hold=False):
    """plot on two axis in a blocking or non blocking mode."""

    ax1, ax2 = add_clickable_markers2(ax=ax1), add_clickable_markers2(ax=ax2, hold=hold)

    print(ax1.markers)
    print(ax2.markers)

    return ax1, ax2


def test_add_clickable_markers2(a, b):

    ax1 = plt.subplot(121)
    plt.imshow(a)
    ax2 = plt.subplot(122)
    plt.imshow(b)

    return test_multiaxis(ax1, ax2)


def test_hold(a):
    """test hold mechanism independently by how it is implemented inside
    add_clickable_marker2.
    test single axis figure and figure with two subplots holding until enter
        key is pressed."""

    print("test on single figure, will advance only after enter is pressed.")
    plt.close('all')
    plt.figure(1)
    plt.imshow(a)
    ax1 = add_clickable_markers2(hold=True)
    print(ax1.markers)

    print("test on subplot (only last one has hold)")
    plt.figure(2)
    ax2 = plt.subplot(121)
    plt.imshow(a)
    ax3 = plt.subplot(122)
    plt.imshow(a)
    
    ax2,ax3=add_clickable_markers2(ax=ax2),add_clickable_markers2(ax=ax3,hold=True)  
    print (ax2.markers)
    print (ax3.markers)
    return ax1,ax2,ax3

def test_disconnect():
    """test access to connection ids from functions in event handling. these
    are not stored anywhere, so they must be stored at event function definition
    from main function body."""
    
    imfile=r'test\input_data\IMG_4509_crop.jpg'
    a=plt.imread(imfile)
    ax1,ax2=test_add_clickable_markers2(a,a) 
    print ("function is called and axis returned, events functions connected to cids.")
        # the body of the function connect cids. event handling functions are defined
        #     and reference directly to cids, are these accessible?
    print("get disconnection of cid1 and 2 by pressing enter with focus on figure,")
        # this disconnected the cids associated
    print("then close the figure (disconnect also the 3rd cid).")
    return ax1,ax2


    
    
    
    return markers1,markers2

if __name__=="__main__":

    def retax():
        """diagnostic routine. return coordinates of all plotted points,
        assuming this are the symbols for the plotted markers."""
        ax=plt.gca()
        return [(a.get_xdata(),a.get_ydata()) for a in ax.lines]


    #from generalPlotting.pickPointList import add_clickable_markers
    imfile=r'test\input_data\IMG_4509_crop.jpg'
    a=plt.imread(imfile)

    print ('add points with CTRL+left button, remove closest with CTRL+right. retax() prints the points plotted on current axes, f.markers print marker list.')
    plt.figure(1)
    f1=test_add_clickable_markers(a)

    plt.figure(2)
    ax1,ax2=test_add_clickable_markers2(a,a)    
    
    



from pyGeneralRoutines.fn_add_subfix import fn_add_subfix
"""read a CMM program file saved as text and convert measured points to a x,y,z file."""
def cmmconvert(filename,startline, step, outfile=None):
    l=open(filename,'r').readlines()
    c=[(a[0].split('<')[-1],a[1],a[2][:-1]) for a in [ll.split(',')[0:3] for ll in l [startline::step]]]
    if outfile is None: outfile=fn_add_subscript(filename,'_converted')
    open(outfile,'w').write('\n'.join(['\t'.join(x) for x in c]))
    

if __name__=='__main__':
    filename='measure_data/2014_11_21/01_OP2S06_surface.txt'
    startline=55
    step=6
    cmmconvert(filename, startline, step)




"""functions operating on a list of Data2D objects"""
# turned into a class derived from list 2020/01/16, no changes to interface,
# everything should work with no changes.

import os
import matplotlib.pyplot as plt
import numpy as np
import pdb

from pySurf.readers.format_reader import auto_reader
from pySurf.data2D import plot_data,get_data, level_data, save_data, rotate_data, remove_nan_frame, resample_data
from pySurf.data2D import read_data,sum_data, subtract_data, projection, crop_data, transpose_data, apply_transform, register_data
from plotting.multiplots import find_grid_size, subplot_grid
from pySurf.psd2d import psd2d,plot_psd2d,psd2d_analysis,plot_rms_power,rms_power
from plotting.backends import maximize
from plotting.add_clickable_markers import add_clickable_markers2

from pySurf.points import matrix_to_points2
from copy import deepcopy
from dataIO.span import span
from dataIO.fn_add_subfix import fn_add_subfix
from pySurf.data2D import projection
from pySurf.data2D_class import Data2D
from pySurf.data2D import levellegendre
from pySurf.affine2D import find_rototrans,find_affine
from pySurf.readers.instrumentReader import fitsWFS_reader

from IPython.display import display
from dataIO.superlist import Superlist
import itertools
from plotting.multiplots import commonscale

## FUNCTIONS ##

# Functions operating on a dlist as Python list of Data2D objects.
# Can be used equally on a Dlist object.

    
def topoints(data,level=None):
    """convert a dlist to single set of points containing all data.
    if level is passed, points are leveled and the value is passed as argument (e.g. level=(2,2) levels sag along two axis)."""
    if level is not None:
        plist = [d.level(level) for d in data] 
    plist = [d.topoints() for d in data]    
    return np.vstack(plist)

def load_dlist(rfiles,reader=None,*args,**kwargs):
    """Extracted from plot_repeat. Read a set of rfiles to a dlist.
    readers and additional arguments can be passed as scalars or lists.

    You can pass additional arguments to the reader in different ways:
     - pass them individually, they will be used for all readers
         load_dlist(.. ,option1='a',option2=1000)
     - to have individual reader parameters pass them as dictionaries (a same number as rfiles),
         load_dlist(.. ,{option1='a',option2=1000},{option1='b',option3='c'},..)
         in this case reader must be explicitly passed (None is acceptable value for auto).

    Example:
        dlist=load_dlist(rfiles,reader=fitsWFS_reader,scale=(-1,-1,1),
                units=['mm','mm','um'])

        dlist2=load_dlist(rfiles,fitsWFS_reader,[{'scale':(-1,-1,1),
                'units':['mm','mm','um']},{'scale':(1,1,-1),
                'units':['mm','mm','um']},{'scale':(-1,-1,1),
                'units':['mm','mm','$\mu$m']}])
    2019/04/08 made function general from plot_repeat, moved to dlist.
    """

    #pdb.set_trace()
    if reader is None:
        #reader=auto_reader(rfiles[0])
        reader = [auto_reader(r) for r in rfiles]
        
    if np.size(reader) ==1:
        reader=[reader]*len(rfiles)
        
    '''
    if kwargs : #passed explicit parameters for all readers
        pdb.set_trace()
        kwargs=[kwargs]*len(rfiles)
    else:
        if np.size(args) ==1:
            kwargs=[args]*len(rfiles)
        else:
            if np.size(args) != len(rfiles):
                raise ValueError
    '''

    if kwargs : #passed explicit parameters for each reader
        # Note, there is ambiguity when rfiles and a kwargs value have same
        # number of elements()
        #pdb.set_trace()
        #vectorize all values
        for k,v in kwargs.items():
            if (np.size(v) == 1):
                kwargs[k]=[v]*len(rfiles)    
            elif (len(v) != len(rfiles)):
                kwargs[k]=[v]*len(rfiles)
            #else:  #non funziona perche' ovviamente anche chiamando esplicitamente, sara'
            #  sempre di lunghezza identica a rfiles.
            #    print ('WARNING: ambiguity detected, it is not possible to determine'+
            #    'if `%s` values are intended as n-element value or n values for each data.\n'+
            #    'To solve, call the function explicitly repeating the value.'%k)
    
    # 2020/07/10 args overwrite kwargs (try to avoid duplicates anyway).
    # args were ignored before.
    
    #if not args:  #assume is correct number of elements
    #    args = [[]]*len(rfiles)   ## Non va fatto cosi'!! senno' duplica "by ref"
    
    #pdb.set_trace()
    
    #transform vectorized kwargs in list of kwargs
    kwargs=[{k:v[i] for k,v in kwargs.items()} for i in np.arange(len(rfiles))]
    
    #kwargs here is a list of dictionaries {option:value}, matching the readers
    #dlist=[Data2D(file=wf1,reader=r,**{**k, **a}) for wf1,r,k,a in zip(rfiles,reader,args,kwargs)]
    dlist=[Data2D(file=wf1,reader=r,*a,**k) for wf1,r,a,k in zip(rfiles,reader,args,kwargs)]
    
    return dlist

def test_load_dlist(rfiles):

    dlist=load_dlist(rfiles,reader=fitsWFS_reader,scale=(-1,-1,1),
            units=['mm','mm','um'])

    dlist2=load_dlist(rfiles,fitsWFS_reader,[{'scale':(-1,-1,1),
            'units':['mm','mm','um']},{'scale':(1,1,-1),
            'units':['mm','mm','um']},{'scale':(-1,-1,1),
            'units':['mm','mm','$\mu$m']}])
    return dlist,dlist2

def plot_data_repeat(dlist,name="",num=None,*args,**kwargs):
    """"given a list of Data2D objects dlist, plots them as subplots on a grid with shared x and y scales in maximized window. colorscale is independent for each subplot.
    returns stats.
    num is the figure number to plot on a specofic figure, other arguments are passed to plot.
    """

    res=[]
    #fig,axes=plt.subplots(1,len(dlist),num=1)
    xs,ys=find_grid_size(len(dlist),3,square=False)
    fig,axes=plt.subplots(xs,ys,num=num,clear=True)
    plt.close() #prevent from showing inline in notebook with %matplotlib inline
    axes=axes.flatten()
    maximize()
    
    for i,(ll,ax) in enumerate(zip(dlist,axes)):
        plt.subplot(xs,ys,i+1,sharex=axes[0],sharey=axes[0])
        ll.plot(stats=True,*args,**kwargs)
        res.append(ll.std())
        #plt.clim([-3,3])
        plt.clim(*(np.nanmean(ll.data)+np.nanstd(ll.data)*np.array([-1,1])))
    plt.suptitle(name+' RAW (plane level)')
    for ax in axes[:len(dlist)-1:-1]:
        fig.delaxes(ax)
        #plt.pause(0.1)
        
    commonscale(plt.gcf())
        
    return res       

def dcouples_plot(dlist,level=True):
    """calculate and plots rotating differences, data are supposed to be already aligned. 
    plots are generated on a grid, x and y axes are shared, color scale is automatic for each subplot.
    Note, data are leveled by default (level=True).
    
    """
    
    dcouples=[c[1]-c[0] for c in list(itertools.combinations(dlist, 2))]
    if level:
        dcouples=[d.level() for d in dcouples]

    plt.clf()
    maximize()
    
    xs,ys=find_grid_size(len(dcouples),square=True)[::-1]
    fig,axes=plt.subplots(xs,ys)
    plt.close() #prevent from showing inline in notebook with %matplotlib inline
    if len(np.shape(axes))>1:
        axes=axes.flatten()
    elif len(np.shape(axes))==0:
        axes=[axes]
    maximize()
    for i,(ll,ax) in enumerate(zip(dcouples,axes)):
        plt.subplot(xs,ys,i+1,sharex=axes[0],sharey=axes[0])
        ll.plot()
        plt.clim(*(ll.std()*np.array([-1,1])))
    for ax in axes[:len(dcouples)-1:-1]:
        fig.delaxes(ax)
        #plt.pause(0.1)
    #plt.tight_layout()
    

    #return [d.std() for d in [diff21,diff31,diff32]]
    return dcouples


def compare_images(datalist, x=None, y=None, fignum=None, titles=None,
                   vmin=None, vmax=None, commonscale=False, axis=0, axmax=0,
                   *args, **kwargs):
    """return a generator that plots n images in a list in subplots with shared
    zoom and axes. datalist is a list of data in format (data, x, y).
    x and y provide plot range (temporarily, ideally want to be able to 
    plot same scale).
    fignum window where to plot, if fignnum is 0 current figure is cleared,
    if None new figure is created.
    axis defines the axis with larger number of plots (default ncols >= nrows)
    axmax maximum nr of plots along shorter dim.
    """
    #modified again 2018/06/05 to accept list of data,x,y triplets. x and y are 
    # accepted as range.
    # old docstring was:
    # return a generator that plots n images in a list in subplots with shared
    # zoom and axes. datalist is a list of 2d data on same x and y.
    # fignum window where to plot, if fignnum is 0 current figure is cleared,
    # if None new figure is created.
    
    
    # this was changed a couple of times in interface, for example in passing
    # ((data,x,y),...) rather than (data1,data2),x,y
    # it can be tricky in a general way if data don't have all same size,
    # at the moment it is assumed they have.
    # in any case x and y are used only to generate the extent,
    # that is then adapted to the size of data.

    gridsize = find_grid_size(len(datalist),axmax)
    if axis == 1: gridsize=gridsize[::-1]

    fig = fignumber(fignum)

    plt.clf()
    ax = None

    # this is to set scale if not fixed
    d1std = [np.nanstd(data[0]) for data in datalist]
    std = min(d1std)

    for i, d in enumerate(datalist):
        """adjust to possible input formats"""
        data, x, y = d
        if x is None:
            x = np.arange(data.shape[1])
        if y is None:
            y = np.arange(data.shape[0])
        ax = plt.subplot(gridsize[0], gridsize[1], i+1, sharex=ax, sharey=ax)
        if titles is not None:
            print("titles is not none, it is ", titles)
            plt.title(titles[i])
        # plt.xlabel('X (mm)')
        # plt.ylabel('Y (mm)')
        s = (std if commonscale else d1std[i])
        d1mean = np.nanmean(data)
        aspect=kwargs.pop('aspect',None)
        axim = plt.imshow(data, extent=np.hstack([span(x),span(y)]),
                          interpolation='none', aspect=aspect,
                          vmin=kwargs.get('vmin', d1mean-s),
                          vmax=kwargs.get('vmax', d1mean+s),
                          *args, **kwargs)
        plt.colorbar()
        yield ax

def multimarkers(datalist):
    """add clickable markers to a list of data using `add_clickable_markers2`. 
    Not sure it is updated, see also Dlist."""
    
    
    plt.figure()
    
    #difficile condividere gli assi visto che non possono essere passati
    s=subplot_grid(len(datalist))
    a=[add_clickable_markers2(s[0])]
    for ss in s[1:-1]:
        plot_data(datalist)
        a.append(add_clickable_markers2(ss))
    a.append(add_clickable_markers2(s[-1],hold=True))
    return a

def mark_data(datalist,outfile=None,deg=1,levelfunc=None,propertyname='markers',direction=0):
    """plot all data in a set of subplots. Allows to interactively put markers
    on each of them. Return list of axis with attached property 'markers'"""
    import logging
    from matplotlib.widgets import MultiCursor
    from plotting.add_clickable_markers import add_clickable_markers2

    try:
        logger
    except NameError:
        logger=logging.getLogger()
        logger.setLevel(logging.DEBUG)
    #datalist = [d() for d in datalist]]
    axes=list(compare_images([[levellegendre(y,wdata,deg=6),x,y] for (wdata,x,y) in datalist],
        commonscale=True,direction=direction))
    fig=plt.gcf()
    multi = MultiCursor(fig.canvas, axes, color='r', lw=.5, horizOn=True, vertOn=True)
    plt.title('6 legendre removed')

    if outfile is not None:
        if os.path.exists(fn_add_subfix(outfile,'_markers','.dat')):
            np.genfromtxt(fn_add_subfix(outfile,'_markers','.dat'))

    for a in axes:
        a.images[0].set_clim(-0.01,0.01)
        add_clickable_markers2(ax=a,propertyname=propertyname) #,block=True)

    #display(plt.gcf())

    #Open interface. if there is a markers file, read it and plot markers.
    #when finished store markers

    for a in axes:
        logger.info('axis: '+str(a))
        logger.info('markers: '+str(a.markers))
        print(a.markers)

    if outfile is not None:
        np.savetxt(fn_add_subfix(outfile,'_markers','.dat'),np.array([a.markers for a in axes]))
        plt.savefig(fn_add_subfix(outfile,'_markers','.png'))

    return axes

def add_markers(dlist):
    """interactively set markers, when ENTER is pressed,
    return markers as list of ndarray.
    It was align_active interactive, returning also trans, this returns only markers,
    transforms can be obtained by e.g. :
    m_trans=find_transform(m,mref) for m in m_arr]
      """

    #set_alignment_markers(tmp)
    xs,ys=find_grid_size(len(dlist),5)[::-1]

    # differently from plt.subplots, axes are returned as list,
    # beacause unused axes were removed (in future if convenient it could
    # make return None for empty axes)
    fig,axes=subplot_grid(len(dlist),(xs,ys),sharex='all',sharey='all')

    #axes=axes.flatten()  #not needed because a list
    maximize()
    for i,(d,ax) in enumerate(zip(dlist,axes)):
        plt.sca(ax)
        #ll=d.level(4,byline=True)
        ll = d
        ll.plot()
        #plt.clim(*remove_outliers(ll.data,nsigma=2,itmax=1,span=True))
        add_clickable_markers2(ax,hold=(i==(len(dlist)-1)))

    return [np.array(ax.markers) for ax in axes]

def align_interactive(dlist,find_transform=find_affine,mref=None):
    """plot a list of Data2D objects on common axis and allow to set
    markers. When ENTER is pressed, return markers and transformations"""

    m_arr = add_markers (dlist)

    #populate array of transforms
    mref = mref if mref is not None else m_arr[0]
    m_trans = [find_transform(m,mref) for m in m_arr]

    return m_arr,m_trans

def psd2an(dlist,wfun=np.hanning,
                ymax=None,outfolder="",subfix='_psd2d',*args,**kwargs):
    """2d psd analysis with threshold on sag removed data. Return a list of psd2d.
    ymax sets top of scale for rms right axis.
    if outfolder is provided, save psd2d_analysis plot with dlist names+subfix"""
    m_psd=[]
    title = kwargs.pop('title','')
    #pdb.set_trace()
    for dd in dlist:
        m_psd.append(dd.level(2,axis=0).psd(analysis=True,
            title=title,wfun=wfun,*args,**kwargs))
        #psd2d_analysis(*dd.level(2,byline=True)(),
        #                 title=os.path.basename(outfolder),wfun=wfun,*args,**kwargs)
        #m_psd.append((fs,ps))
        ax=plt.gcf().axes[-1]
        ax.set_ylim([0,ymax])
        plt.grid(0)
        ax.grid()
        plt.suptitle(dd.name+' - hanning window - sag removed by line')
        if outfolder is not None:
            plt.savefig(os.path.join(outfolder,dd.name+subfix+'.png'))
        #pr=projection(pl[0].data,axis=1,span=1)
        #plot_psd(pl[0].y,pr[0],label='avg')
        #plot_psd(pl[0].y,pr[1],label='min')
        #plot_psd(pl[0].y,pr[2],label='max')
    return m_psd

def extract_psd(dlist,rmsthr=0.07,rmsrange=None,prange=None,ax2f=None,dis=False):
    """use psd2an, from surface files return linear psd."""
    m_psd=psd2an(dlist,rmsthr=rmsthr,rmsrange=rmsrange,prange=prange,ax2f=ax2f,dis=dis)

    plt.figure()
    plt.clf()
    labels=[d.name for d in dlist]
    m_tot=[]
    #gives error, m_psd is a list of PSD2D objects, there is no f,p, check psd2an
    for P,l in zip(m_psd,labels):
        (f,p) = P()
        ptot=projection(p,axis=1)
        #plot_psd(f,ptot,label=l,units=d.units)
        #np.savetxt(fn_add_subfix(P.name,'_calib_psd','.dat',
        #    pre=outfolder+os.path.sep),np.vstack([fs,ptot]).T,
        #    header='SpatialFrequency(%s^-1)\tPSD(%s^2 %s)'%(u[0],u[2],u[0]))
        m_tot.append((f,ptot))
    # plt.title(outname)

    plt.legend(loc=0)
    if dis:
        display(plt.gcf())
    return m_tot

def psd2d(dlist,ymax=None,subfix='_psd2d',*args,**kwargs):
    """2d psd analysis of a dlist. 
    Doesn't do any additional processing or plotting (must be done externally)
    [`outfolder` is being removed].
    Any parameter for `Data2D.psd` are accepted,
    however, parameters are not vectorized (must be the same for all data).
    [see example of vectorization e.g. in `load_dlist`] 
    Return a list of psd2d.
    
    from `psd2an`
    ymax sets top of scale for rms right axis.
    if outfolder is provided, save psd2d_analysis plot with dlist names+subfix"""
    
    m_psd=[]
    title = kwargs.pop('title','')
    #pdb.set_trace()
    for dd in dlist:
        m_psd.append(dd.psd(analysis=True,
            title=title,wfun=wfun,*args,**kwargs))
        #psd2d_analysis(*dd.level(2,byline=True)(),
        #                 title=os.path.basename(outfolder),wfun=wfun,*args,**kwargs)
        #m_psd.append((fs,ps))
        ax=plt.gcf().axes[-1]
        ax.set_ylim([0,ymax])
        plt.grid(0)
        ax.grid()
        plt.suptitle(dd.name+' - hanning window - sag removed by line')
        if outfolder is not None:
            plt.savefig(os.path.join(outfolder,dd.name+subfix+'.png'))
        #pr=projection(pl[0].data,axis=1,span=1)
        #plot_psd(pl[0].y,pr[0],label='avg')
        #plot_psd(pl[0].y,pr[1],label='min')
        #plot_psd(pl[0].y,pr[2],label='max')
    return m_psd

## CLASS ##

# Minimal implementation, should broadcast properties

class Dlist(Superlist):
    """A list of pySurf.Data2D objects on which unknown operations are performed serially."""           
    
    def topoints(self,level=True):
        """convert a dlist to single set of points containing all data."""
        plist = topoints(self.data,level = None)
        return plist  


import matplotlib.pyplot as plt
import numpy as np


def plot_labels(labels, xdata, ydata, xoffset=None, yoffset=None,plain=False):
    """"Add fancy labels with arrows pointing at xdata,ydata.
    offsets can be passed as offsets in data coordinates as scalar or vectors.
    default is a fraction of axes, or can use """
    
    #set offsets format as vectors
    frac=1/20. #fraction of full axis range used as default offset 
    xyc='data' #'offset points'
    if xoffset is None:
        #xyc = 'axes fraction'
        #xoffset = - frac
        xoffset = -(plt.xlim()[1]-plt.xlim()[0])*frac
    if yoffset is None:
        #xyc = 'axes fraction'
        #yoffset = frac
        yoffset = (plt.ylim()[1]-plt.ylim()[0])*frac
    if np.size(xoffset)==1:
        xoffset=np.repeat(xoffset,len(labels))
    if np.size(yoffset)==1:
        yoffset=np.repeat(yoffset,len(labels))
    
    for label, x, y, xo, yo in zip(labels, xdata, ydata, xoffset, yoffset):
        if plain:
            plt.text(x+xo, y+yo, label)
        else:
            plt.annotate(
                label, 
                xy = (x, y), xytext = (x+xo, y+yo), 
                textcoords = xyc, ha = 'right', va = 'bottom',
                bbox = dict(boxstyle = 'round,pad=0.5', fc = 'yellow', alpha = 0.5),
                arrowprops = dict(arrowstyle = '->', connectionstyle = 'arc3,rad=0'))
            
if __name__=="__main__":
    #this is what I want to pass as label, it must be a 3 element list with [string, xpos, ypos]
    xdata=(1.,2.,3.)
    ydata=(1.,4.,9.)
    labels=('uno','due','tre')
    
    plt.figure(1)    
    plt.plot(xdata,ydata,'o')
    plt.plot(xdata,ydata)
    plot_labels(labels, xdata, ydata, plain=True)
    plt.show()
    
    plt.figure(2)
    plt.plot(xdata,ydata,'o')
    plt.plot(xdata,ydata)
    plot_labels(labels, xdata, ydata, plain=True)
    plt.show()